
"""Intercept requests to ensure that HTTPS is reported correctly.

"""

from __future__ import unicode_literals

from .urllib3 import intercept_urllib3


def intercept_requests(agent, mod):
    #print("intercept_requests")
    # requests ships with its own version of urllib3, so we need to manually intercept it.
    intercept_urllib3(agent, mod.packages.urllib3)



"""Interceptor for Bottle.

"""

import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor


class BottleInterceptor(BaseInterceptor):
    def add_exception(self, func, *args, **kwargs):
        self.agent.logger.info('Modulename: BottleInterceptor class')
        with self.log_exceptions():
            bt = self.bt
            self.agent.logger.info("Modulename: BottleInterceptor class inside add_exception function bt value is :{0}".format(bt))
            if bt:
                bt.add_exception(*sys.exc_info())
        return func(*args, **kwargs)



def intercept_bottle(agent, mod):
    WSGIInterceptor(agent, mod.Bottle).attach('__call__')
    #print("framework.bottle.intercept_bottle: intercepting bottle ......")
    BottleInterceptor(agent, mod.HTTPError).attach('__init__', patched_method_name='add_exception')





"""Intercept the MySQLdb package.

"""

from __future__ import unicode_literals
import weakref

from .dbapi import DbAPIConnectionInterceptor, DbAPICursorInterceptor


class MySQLdbConnectionInterceptor(DbAPIConnectionInterceptor):
    def get_backend_properties(self, conn, *args, **kwargs):
        self.agent.logger.info('Modulename: MySQLdbCursorInterceptor  class inside get_backend_properties function')
        host = kwargs.get('host', 'localhost')
        port = kwargs.get('port', '3306')
        db = kwargs.get('database', kwargs.get('db', ''))
        return host, port, db, 'MYSQL'


class MySQLdbCursorInterceptor(DbAPICursorInterceptor):
    def get_connection(self, cursor):
        self.agent.logger.info('Modulename: MySQLdbCursorInterceptor  class inside get_connection function ') 
        conn = cursor.connection
        self.agent.logger.info("Modulename: MySQLdbCursorInterceptor  class inside get_connection function connection is {0} ".format(conn))

        if isinstance(conn, weakref.ReferenceType):
            conn = conn()
        return conn


def intercept_MySQLdb_connection(agent, mod):
    MySQLdbConnectionInterceptor(agent, mod.Connection, MySQLdbCursorInterceptor).attach('__init__')



from pyclbr import readmodule
import os
import threading 
    
    
def application_callable_pro(self, application, instance, environ, start_response):
    stack = inspect.stack()
    temp={}
    the_method_called = stack[3][0].f_code.co_name
    request = Request(environ)
    prof = cProfile.Profile()
    prof.enable()
    bt = self.start_business_transaction(request.full_path, '')
    self.agent.logger.debug("Modulename: WSGI intercepto class || bt value is :{0}".format(bt))
    method = self.agent.method_entry(bt,the_method_called)

    try:
        self.agent.logger.info("Modulename: WSGI interceptor class || request is :{0} || request path is {1} || query_string: {2} || host: {3} ".format(request,(request.path,request.query_string,request.host)))
    except:
        pass

    try:
        response = application(instance, environ, self._make_start_response_wrapper(start_response))
        self.agent.logger.info("Modulename: WSGI interceptor class || response is :{0}".format(response))

    except:
        with self.log_exceptions():
            if bt:
                bt.add_exception(*sys.exc_info())
        raise
    finally:
        method = the_method_called
        self.agent.method_exit(bt,method)
        self.end_business_transaction(bt)
        prof.disable()
        prof.print_stats()
        p = pstats.Stats(prof)
        #client application path
        client_app_root = os.path.dirname(sys.argv[0])

        for k,v in p.stats.items():

            if client_app_root in k[0]:
                module = k[0]
                print("module:-", k[0])
                module_path = os.path.split(os.path.abspath(module))[0]
                modulefullname  = os.path.split(os.path.abspath(module))[1]
                module_name = modulefullname.split('.')[0]
                methodname = k[-1]
                print('module path:-', module_path)
                print('module name:-', module_name)
                print('method name:- ',k[-1])
                print('Total call:- ',v[1])
                print('Response time of method:-',v[3])
                mod = readmodule(module_name, path=[module_path])
                li = []
                for k, v in mod.items():
                    li.append(k)
                    methods = v.methods.items()
                    lk = []
                    for method, lineno in methods:
                        lk.append(method)
                    li.append(lk)
                count=0
                for i in li:
                    count +=1
                    if type(i)==str:
                        pass
                    else:
                        for j in i:
                            if j == methodname:
                                print("classname:-",li[count-2])
                            else:
                                pass
    return response

def application_callable(self, application, instance, environ, start_response):
    stack = inspect.stack()
    temp={}
    the_method_called = stack[3][0].f_code.co_name
    request = Request(environ)
    bt = self.start_business_transaction(request.full_path, '')
    self.agent.logger.debug("Modulename: WSGI interceptor class || bt value is :{0}".format(bt))
    method = self.agent.method_entry(bt,the_method_called)

    try:
        self.agent.logger.info("Modulename: WSGI interceptor class || request is :{0} || request path is {1} || query_string: {2} || host: {3} ".format(request,(request.path,request.query_string,request.host)))
    except:
        pass

    try:
        response = application(instance, environ, self._make_start_response_wrapper(start_response))
        self.agent.logger.info("Modulename: WSGI interceptor class || response is :{0}".format(response))

    except:
        with self.log_exceptions():
            if bt:
                bt.add_exception(*sys.exc_info())
        raise
    finally:
        method = the_method_called
        self.agent.method_exit(bt,method)
        self.end_business_transaction(bt)

    return response



from . import frameworks, logging, http, sql,Instrumentation
import sys

try:
    from importlib.machinery import ModuleSpec
except ImportError:
    pass

try:
    from importlib import reload
except ImportError:
    from imp import reload


URL_PROPERTY_MAX_LEN = 100
HOST_PROPERTY_MAX_LEN = 50
DB_NAME_PROPERTY_MAX_LEN = 50
VENDOR_PROPERTY_MAX_LEN = 50


#from pythonagent.agent.internal.intercept_module import BT
#BT_INTERCEPTORS = BT




# BT_INTERCEPTORS = (
#     # Entry points
#     ('flask', frameworks.intercept_flask),
#     ('django.core.handlers.wsgi', frameworks.intercept_django_wsgi_handler),
#     ('django.core.handlers.base', frameworks.intercept_django_base_handler),
#
#     # HTTP exit calls
#     ('httplib', http.intercept_httplib),
#     ('http.client', http.intercept_httplib),
#     ('urllib3', http.intercept_urllib3),
#     ('requests', http.intercept_requests),
#
#     # Logging
#     ('logging', logging.intercept_logging),
#
#     # SQL exit calls
#     ('psycopg2', sql.psycopg2.intercept_psycopg2_connection),
#     ('pymysql.connections', sql.pymysql.intercept_pymysql_connections),
#     ('mysql.connector.connection', sql.mysql_connector.intercept_mysql_connector_connection),
#     ('MySQLdb.connections', sql.mysqldb.intercept_MySQLdb_connection),
# )


def add_hook(agent):
    """Add the module interceptor hook for pythonagent, if it's not already registered.

    """

    interceptor = ModuleInterceptor(agent)
    sys.meta_path.insert(0, interceptor)
    return interceptor


class ModuleInterceptor(object):
    """Intercepts finding and loading modules in order to monkey patch them on load.

    """

    def __init__(self, agent):
        super(ModuleInterceptor, self).__init__()
        self.agent = agent
        self.module_hooks = {}
        self.intercepted_modules = set()

    def find_spec(self, full_name, path, target=None):
        if full_name in self.module_hooks:
            return ModuleSpec(full_name, self)
        return None

    def find_module(self, full_name, path=None):
        #self.agent.logger.info("Finding module {0}".format(full_name))
        self.agent.logger.info('Modulename: ModuleInterceptor class || full_name {0}'.format(full_name))
        if full_name in self.module_hooks:
            return self
        return None

    def load_module(self, name):
        # Remove the module from the list of hooks so that we never see it again.
        #self.agent.logger.info('Modulename: ModuleInterceptor class || function name {0}'.format(name))
        #self.agent.logger.info("Loader loading load module {0}".format(name))
        # for module in custom_methods_to_instrument_json["modules"]:
        #   if name == module["moduleName"]:
        #       hooks = Instrumentation.intercept_instrumented_method

        # hooks = self.module_hooks.pop(name, [])
        # print("hooks for module: ",name," ->",hooks)

        #if name in all_module_names:
         #   hooks = Instrumentation.intercept_instrumented_method
         #   all_module_names.remove(name)
        #else:
        #    hooks = self.module_hooks.pop(name, [])
 
        hooks = self.module_hooks.pop(name, [])
        #self.agent.logger.info("hooks for module: {0} and hook {1}".format(name,hooks))
        #self.agent.logger.info('Modulename: ModuleInterceptor class || hooks for module {0}'.format(hooks))
        if name in sys.modules:
            # Already been loaded. Return it as is.
            return sys.modules[name]

        #self.agent.logger.debug('Intercepting import %s', name)
        #self.agent.logger.info(("Intercepting import {0}".format(name)))
        #self.agent.logger.info('Modulename: ModuleInterceptor class || intercepting import {0}'.format(name))
        __import__(name)  # __import__('a.b.c') returns <module a>, not <module a.b.c>
        module = sys.modules[name]  # ...so get <module a.b.c> from sys.modules

        self._intercept_module(module, hooks)

        return module

    def call_on_import(self, module_name, cb):
        #print("CALL ON IMPORT")
        #print("module_name, cb", module_name, cb)

        #print("agent/probes/init.call_on_import() ...")
        #self.agent.logger.info('Modulename: ModuleInterceptor class || module_name is {0} and callback is {1}'.format(module_name,cb))
        #self.agent.logger.info("modulename: %s", module_name, " callback: %s", cb)
        #self.agent.logger.info('Modulename: ModuleInterceptor class inside call_on_import  function callback is {0}'.format(cb))

        # print(custom_methods_to_instrument_json["modules"])
        # print("hooks for module: ",name," ->",hooks)
        
        #if module_name in all_module_names:
        #    hooks = Instrumentation.intercept_instrumented_method
        #    self._intercept_module(sys.modules[module_name], [hooks])
        
        if module_name in sys.modules:
            #print("FOUND IN SYS MODULES")
            #print("module" ,module_name, " found in sys.modules. Calling intercept module")
            self._intercept_module(sys.modules[module_name], [cb])
        else:
            #print("NOT FOUND IN SYS MODULES")
            #print("module" ,module_name, " not found in sys.modules. Will be called from load modules")
            self.module_hooks.setdefault(module_name, [])
            self.module_hooks[module_name].append(cb)
            #print("MODULE HOOKS", self.module_hooks)


        #print('self.module_hooks value',self.module_hooks)
        #if module_name in all_module_names:
         #   hooks = Instrumentation.intercept_instrumented_method
         #   __import__(module_name)
         #   self._intercept_module(sys.modules[module_name], [hooks])

    def _intercept_module(self, module, hooks):
        try:
            for hook in hooks:
                #self.agent.logger.debug("Running {0} hook {1}".format(module.__name__, hook))
                hook(self.agent, module)
            self.intercepted_modules.add(module)
        except:
            self.agent.logger.exception("Exception in {0} hook.".format(module.__name__))

            # Re-import to ensure the module hasn't been partially patched.
            self.agent.logger.debug("Re-importing {0} after error in module hook".format( module.__name__))
            reload(module)






""""
Local Testing for Python Agent on AWS Lambda
"""

import sys
import os

python_agent_parent_dir = os.path.abspath(os.path.join(__file__, "../.."))
sys.path.append(python_agent_parent_dir)

from pythonagent.bootstrap.cavagent_lambda_wrapper import handler


if __name__ == "__main__":

    event = {"path": "test_local_application"}

    class Context:
        def __init__(self):
            self.log_stream_name = "test_request_id_localapp"
            self.aws_request_id = "test_request_id_localapp"
            self.function_name = "test_function_localapp"

    context = Context()

    response = handler(event, context)




from __future__ import unicode_literals

import copy
import logging
import logging.handlers
import os.path
import re
from datetime import datetime
import sys
from pythonagent import config
from pythonagent.lib import default_log_formatter, mkdir

LOGGING_MAX_BYTES = 10 * 1024 * 1024
LOGGING_MAX_NUM_FILES = 5


def get_log_level():
    default_logging_level = logging.WARNING
    allowed_logging_levels = {'DEBUG': logging.DEBUG,
                              'INFO': logging.INFO,
                              'WARNING': logging.WARNING,
                              'ERROR': logging.ERROR,         # Added for testing
                              'CRITICAL': logging.CRITICAL    # Added for testing
                              }

    level = config.LOGGING_LEVEL.upper()
    ret_val = allowed_logging_levels.get(level, default_logging_level)
    # NOTSET=0
    # DEBUG=10
    # INFO=20
    # WARN=30
    # ERROR=40
    # CRITICAL=50
    return allowed_logging_levels.get(level, default_logging_level)


def get_log_filename():
    non_alphanumeric = re.compile(r'\W+')
    sanitize = lambda x: non_alphanumeric.sub('_', x)
    filename = '-'.join(
        map(sanitize, [config.APP_NAME, config.TIER, config.SERVER, '{:%Y-%m-%d}'.format(datetime.now())])) + '.log'
    return os.path.join(config.LOGS_DIR, filename)


def debug_enabled(logger):
    for handler in logger.handlers:
        if handler.__class__ is logging.StreamHandler:
            return True
    return False


def enable_debug(logger):
    logger.setLevel(logging.DEBUG)

    for handler in logger.handlers:
        handler.setLevel(logging.DEBUG)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(default_log_formatter)
    logger.addHandler(stream_handler)


def disable_debug(logger):
    level = get_log_level()
    logger.setLevel(level)

    handlers = copy.copy(logger.handlers)
    logger.handlers = []
    for handler in handlers:
        if handler.__class__ is not logging.StreamHandler:
            handler.setLevel(level)
            logger.addHandler(handler)


def configure_logging():
    try:
        logger = logging.getLogger('pythonagent.agent')
        level = get_log_level()
        print("Logging level:", level)
        logger.setLevel(level)
        path = get_log_filename()

        print("Logging file path name", path)

        agent_env = None

        if "CAV_APP_AGENT_Env" in os.environ:
            if os.environ['CAV_APP_AGENT_Env'] == "AWS_LAMBDA":
                agent_env = "AWS_LAMBDA"
            elif os.environ['CAV_APP_AGENT_Env'] == "NATIVE":
                agent_env = "NATIVE"
        else:
            agent_env = "AWS_LAMBDA"

        if agent_env == "AWS_LAMBDA":
            consolHandler = logging.StreamHandler(sys.stdout)
            consolHandler.setLevel(level)
            consolHandler.setFormatter(default_log_formatter)
            logger.addHandler(consolHandler)

        else:
            mkdir(os.path.dirname(path))
            rotating_file_handler = logging.handlers.RotatingFileHandler(path, maxBytes=LOGGING_MAX_BYTES,
                                                                         backupCount=LOGGING_MAX_NUM_FILES - 1)
            # logging.getLogger('pythonagent.agent').addHandler()
            rotating_file_handler.setLevel(level)
            rotating_file_handler.setFormatter(default_log_formatter)
            logger.addHandler(rotating_file_handler)

        if config.DEBUG_LOG:
            enable_debug(logger)

        has_handlers = logger.hasHandlers()
        logger.warning("has handlers {}".format(str(has_handlers)))
        logger.propagate = False
        logger.warning("has handlers {}".format(str(has_handlers)))


    except:
        logging.getLogger('pythonagent.agent').exception('Logging configuration failed.')



from __future__ import unicode_literals

from pythonagent.agent.probes.frameworks.wsgi import WSGIMiddleware
#from pythonagent import config
#config.WSGI_SCRIPT_ALIAS = input("Enter the absolute path of WSGI application script. (command to get 'readlink -f <script_path.py>') : ")
application = WSGIMiddleware()



# Copyright (c) cavisson, Inc., and its affiliates
# 2015
# All Rights Reserved

import os
import sys

import sys,os
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname( __file__ ))))


from lang import keys
from main.pytrace import CommandInvocationError, CommandExecutionError, parse_options, HELP_OPTIONS
from main.pytrace.commands import run,auto_discovery

pytrace = os.path.basename(sys.argv[0])


COMMANDS = {
    'run': run,
    'auto_discovery': auto_discovery,
}
COMMANDS_ORDER = ["run"]

HELP_TEMPLATE = """\
USAGE
    {pytrace} {usage}

ABOUT
    {about}"""

USAGE = """\
USAGE
    {pytrace} <command> <args...>

ABOUT
    Cavisson Python Agent management utility

COMMANDS
    help                Print this message
    help <command>      Show detailed help for the given pytrace command
"""


def main():
    if len(sys.argv) < 2:
        print_usage()


    command = sys.argv[1]
    args = sys.argv[2:]

    if command in HELP_OPTIONS or command == 'help':
        if args:
            print_usage(command=args[0])
        else:
            print_usage()

    if command not in COMMANDS:
        print_usage(error="%s: unrecognized command" % command)

    mod = COMMANDS[command]

    try:
        options, args = parse_options(getattr(mod, 'OPTIONS', {}), args)

        if 'help' in options:
            print_usage(command=command)

        mod.command(options, args)

    except CommandInvocationError as exc:
        print_usage(command=command, error=str(exc))
        #print("wrong usage!")
    except CommandExecutionError as exc:
        print(str(exc))


def indent(text, spaces=4):
    indentation = ' ' * spaces
    indented_text = ('\n%s' % indentation).join(text.split('\n'))
    return indented_text

def print_usage(command=None, error=None):
    if command and command not in COMMANDS:
        error = '%s: unrecognized command' % command
        command = None

    if command:
        mod = COMMANDS[command]
        about = indent(mod.ABOUT.format(pytrace=pytrace))

        print(HELP_TEMPLATE.format(pytrace=pytrace, usage=mod.USAGE, about=about))

        if getattr(mod, 'OPTIONS', None):
            print('OPTIONS')
            for opt in sorted(keys(mod.OPTIONS)):
                opt_descr = mod.OPTIONS[opt]
                if isinstance(opt_descr, dict):
                    if 'short' in opt_descr:
                        opt = '%s / -%s' % (opt, opt_descr['short'])
                    if opt_descr.get('value', False):
                        opt += ' %s' % (opt_descr.get('value_help', ' <value>'))

                    opt_help = opt_descr.get('help', False)

                    if opt_help is False:
                        continue

                    if opt_descr.get('required', False):
                        opt_help = 'REQUIRED: %s' % opt_help
                else:
                    opt_help = opt_descr

                print('    --%-28s  %s' % (opt, opt_help))
    else:
        print(USAGE.format(pytrace=pytrace))

        for cmd in COMMANDS_ORDER:
            mod = COMMANDS[cmd]
            first_line = mod.ABOUT.splitlines()[0]
            print("    %-18s  %s" % (cmd, first_line))

    if error:
        prefix = "{pytrace} {command}".format(pytrace=pytrace, command=command) if command else pytrace
        print('')
        print("ERROR: {prefix}: {error}".format(prefix=prefix, error=error))

    sys.exit(1 if error else 0)


if __name__ == '__main__':
    main()



from __future__ import unicode_literals
#import sys

from ..base import ExitCallInterceptor
#import abc
#import agent
#from agent.internal.proxy import *
from pythonagent.lang import str


EXIT_CUSTOM = 1
EXIT_SUBTYPE_MONGODB = 'Mongo DB'

def intercept_pymongo(agent, mod):
    class ExitCallListener(mod.monitoring.CommandListener):
        #agent.logger.info("going to get proxy instance (pymongo ExitCallListener Interceptor)")
        #proxy = Proxy.getInstance()
        #print("proxy handle = ", proxy)

        backend_name_format_string = '{HOST}:{PORT} - {DATABASE}'

        def __init__(self):
            self.interceptor = ExitCallInterceptor(agent, None)
            self.exit_call_map = {}

        def get_backend(self, connection_id, database_name):
            #agent.logger.debug('Modulenameintercept_pymongo  class inside get_backend function') 
            print("connection iddddddddddddddddddddddddddddddddd::::::::::::::::::", connection_id)
            host, port = connection_id
            backend_properties = {
                'HOST': host,
                'PORT': str(port),
                'DATABASE': database_name,
            }
            agent.logger.info("Modulenameintercept_pymongo  class || backend_properties is :{0}".format(backend_properties))
            # backend = get_backend(EXIT_CUSTOM, EXIT_SUBTYPE_MONGODB,backend_properties,self.backend_name_format_string)
            return backend_properties
            #return None

        def started(self, event):
            agent.logger.info('Modulenameintercept_pymongo  class inside started  function  is ')
            with self.interceptor.log_exceptions():
                agent.logger.info("inside _execute funciton")
                agent.logger.info("event:{0} ".format(event))
            #    self.agent.logger.info('Modulenameintercept_pymongo  class inside started  function event is'.format(event))
                bt = self.interceptor.bt
                #bt = self.agent.bt
             #   self.agent.logger.info('Modulenameintercept_pymongo  class inside started  function bt is'.format(bt))

                agent.logger.info("bt statment....{0}".format(bt))
                if not bt:
                      bt = agent.get_current_bt()
                agent.logger.info("new bt value{0}".format(bt))
                if bt:
                    backend = self.get_backend(event.connection_id, event.database_name)
                    agent.logger.info("backend..........{0}".format(backend))
                    # backend = True
                    agent.logger.info("Modulenameintercept_pymongo class ||  backend is{0}".format(backend))
                    if backend:
                        # print("connection.host:", backend.dsn)
                        #print("backend: ", backend.dsn)
             #           self.agent.logger.info('Modulenameintercept_pymongo  class inside started function event.command is'.format(event.command))
                        agent.logger.info("host.....".format(backend["HOST"]))
              #          self.agent.logger.info('Modulenameintercept_pymongo  class inside started function  backend["HOST"] is'.format( backend["HOST"]))
                        if backend and not hasattr(event, '_pythonagent_exit_call'):
                            #exit_call =self.proxy.db_call_begin(bt,backend["HOST"] , str(event.command))
                            exit_call = self.interceptor.db_call_begin(bt, backend["HOST"], str(event.command))
                            # exit_call = self.start_exit_call(bt, backend, operation=operation)
                            # event._pythonagent_exit_call = exit_call
               #             self.agent.logger.info('Modulenameintercept_pymongo  class inside started function exit_call is'.format(exit_call))
                        #exit_call = self.interceptor.start_exit_call(bt, backend, str(event.command))
                        self.exit_call_map[event.operation_id] = exit_call

        def succeeded(self, event):
            # proxy.db_call_end(self.exit_call_map.pop(event.operation_id, None))
            #self.interceptor.db_call_end(self.exit_call_map.pop(event.operation_id, None))
            self.interceptor.db_call_end(self.interceptor.bt,self.exit_call_map.pop(event.operation_id,None))
        def failed(self, event):
            self.interceptor.db_call_end(self.interceptor.bt,self.exit_call_map.pop(event.operation_id,None))

    mod.monitoring.register(ExitCallListener())





from __future__ import unicode_literals
import logging
import os
import configparser
cofig= configparser.RawConfigParser()
import sys

from configparser import SafeConfigParser, ConfigParser

try:
    ospath = os.environ.get('ND_HOME')
    path = ospath + '/config/ndsettings.conf'
except Exception as e:
    print("ND HOME NOT SET")
    raise e

#print("path of  NDC config : ", path)

#configParser.read(path)
#path_for_insProfile = configParser.get('PYTHON_AGENT', 'path_profile')

myvars = {}
with open(path) as myfile:
    for line in myfile:
        name, var = line.partition("=")[::2]
        myvars[name.strip()] = var
        #print("myvars----------------------->",myvars)
#TIER = myvars['tier'].strip('\n')
#SERVER = myvars['server'].strip('\n')




int_or_none = lambda v: int(v) if v != '' else None
on_off = lambda v: v.lower() in ('on', 'true', 'yes', 'y', 't', '1')
comma_seperated_list = lambda v: v.replace(' ', '').split(',')

keys = lambda d: d.keys()
values = lambda d: d.values()
# Configuration Options

_CONFIG_OPTIONS_BY_SECTION = {

    'agent': {
        'app': ('APP_NAME', None),
        'tier': ('TIER_NAME', None),
        'node': ('NODE_NAME', None),
        'dir': ('DIR', None),
    },

    'wsgi': {
        'script': ('WSGI_SCRIPT_ALIAS', None),
        'callable': ('WSGI_CALLABLE_OBJECT', None),
        'module': ('WSGI_MODULE', None),
    },

    'log': {
        'dir': ('LOGS_DIR', None),
        'level': ('LOGGING_LEVEL', None),
        'debugging': ('DEBUG_LOG', on_off)
    },

    'proxy': {
        'proxymode': ('PROXY_MODE', None),
        'proxycontype': ('PROXY_CON_TYPE', None),
        'proxyip': ('PROXY_IP', None),
        'proxyport': ('PROXY_PORT', None),
        'sopath': ('SO_PATH', None),
       },
}


# Defaults ###########


CONFIG_FILE = ''

# Agent
APP_NAME = 'CavAgent'
#TIER_NAME = 'test_tier'
#NODE_NAME = 'test'
DIR = ospath + '/CavAgent'
TIER = myvars['tier'].strip('\n')
SERVER = myvars['server'].strip('\n')


# Logging
LOGS_DIR = ospath+'/CavAgent/logs'
print("LOGS_DIR", LOGS_DIR)

if "CAV_APP_AGENT_LoggingLevel" in os.environ:
    LOGGING_LEVEL = os.environ.get('CAV_APP_AGENT_LoggingLevel')
else:
    LOGGING_LEVEL = "WARNING"

DEBUG_LOG = False


# WSGI
WSGI_MODULE = ''
#WSGI_SCRIPT_ALIAS = '/home/cavisson/shop/my-shop/wsgi_example_app.py'
#WSGI_SCRIPT_ALIAS = input("Enter the absolute path of WSGI application script. (command to get 'readlink -f <script_path.py>') : ")
WSGI_SCRIPT_ALIAS = ''
WSGI_CALLABLE_OBJECT = ''

# Controller
CONTROLLER_HOST = '10.20.0.85'
CONTROLLER_PORT = '7892'
SSL_ENABLED = False
ACCOUNT_NAME = ''
ACCOUNT_ACCESS_KEY = ''
HTTP_PROXY_HOST = '10.20.0.79'
HTTP_PROXY_PORT = '1112'
HTTP_PROXY_USER = ''
HTTP_PROXY_PASSWORD_FILE = ''
'''
# Proxy Runtime
PROXY_RUN_DIR = ''

# Proxy Control Service
PROXY_CONTROL_PATH = ''
PROXY_STARTUP_READ_TIMEOUT_MS = 2000
PROXY_STARTUP_INITIAL_RETRY_DELAY_MS = 5000
PROXY_STARTUP_MAX_RETRY_DELAY_MS = 300000

# Config Service
PROXY_CONFIG_SOCKET_NAME = '0'
CONFIG_SERVICE_RELOAD_INTERVAL_MS = 5000
CONFIG_SERVICE_MAX_RETRIES = 3

# Transaction Service
PROXY_INFO_SOCKET_NAME = '0'
PROXY_REPORTING_SOCKET_NAME = '1'
BT_INFO_REQUEST_TIMEOUT_MS = 100

# Snapshot Service
INCLUDE_AGENT_FRAMES = False
SNAPSHOT_PROFILER_INTERVAL_MS = 10
EXIT_CALL_DETAILS_LENGTH = 100
FORCED_SNAPSHOT_INTERVAL = 10

# Transaction Monitor Service
BT_MAX_DURATION_MS = 30 * 60 * 1000
BT_ABANDON_THRESHOLD_MULTIPLIER = 6

# EUM
EUM_DISABLE_COOKIE = False
EUM_USER_AGENT_WHITELIST = ['Mozilla', 'Opera', 'WebKit', 'Nokia']
'''
#RUNNING_MODE = 0

def validate_config(config):
  
    return True


def parse_environ(environ=None, prefix='CAV_'):

    logger = logging.getLogger('pythonagent.agent')
    environ = environ if environ is not None else os.environ

    config = {}
    config_file = environ.get('CAV_CONFIG_FILE')

    if config_file:
        config = parse_config_file(config_file)

    #environ['SO_PATH'] = config['SO_PATH']
    option_descrs = {}

    for options in values(_CONFIG_OPTIONS_BY_SECTION):
        for name, handler in values(options):
            option_descrs[prefix + name] = (name, handler)

    for option in keys(environ):
        if option not in option_descrs:
            continue

        name, handler = option_descrs[option]

        try:
            value = environ[option]
            if handler:
                value = handler(value)
            config[name] = value
        except:
            logger.exception('ignoring %s from environment, parsing value caused exception', option)

    return config


def parse_config_file(filename):
    return None
    logger = logging.getLogger('pythonagent.agent')

    try:
        config = {}
        parser = ConfigParser()

        with open(filename) as fp:
            parser.readfp(fp)

        for section_name in parser.sections():
            try:
                options_map = _CONFIG_OPTIONS_BY_SECTION[section_name]
            except KeyError:  # Unknown section
                logger.warning('%s: skipping unrecognized section [%s]', filename, section_name)
                continue

            for option_name in parser.options(section_name):
                try:
                    env_name, handler = options_map[option_name]

                    value = parser.get(section_name, option_name)
                    if handler:
                        value = handler(value)

                    config[env_name] = value
                except KeyError:  # Unknown option
                    logger.warning('%s: skipping unrecognized option %r of section [%s]',
                                   filename, option_name, section_name)
                except:  # Other errors
                    logger.exception('%s: parsing value for option %r of section [%s] raised exception',
                                     filename, option_name, section_name)

        return config
    except:
        logger.exception('Parsing config file failed.')


def merge(config):

    #mod = globals()
    #mod.update(config)
    update_computed_defaults()


def update_computed_defaults():
    global LOGS_DIR, PROXY_CONTROL_PATH, PROXY_RUN_DIR

    PROXY_RUN_DIR = os.path.join(DIR, 'run')

    #if not PROXY_CONTROL_PATH:
    #    PROXY_CONTROL_PATH = os.path.join(PROXY_RUN_DIR, 'comm')

    if not LOGS_DIR:
        LOGS_DIR = os.path.join(DIR, 'logs')




from pythonagent import config
from pythonagent.agent.internal.logs import configure_logging
from pythonagent.agent.probes.Instrumentation.parser import create_method_array,create_method_json
from pythonagent.agent.probes.Instrumentation.parser import create_method_array,create_method_json
import os
from pythonagent.lang import reload

'''
import os
path = os.path.dirname(config.__file__)
#print("Module path", path)
#os.chdir(path)
#print("Current working dir ==========================================",path)

f = open(os.path.join(path, 'agent/internal/intercept_module.py'), "a+")
f.truncate(0)
#cwd = os.getcwd()  # Get the current working directory (cwd)
#files = os.listdir(cwd)  # Get all the files in that directory
#print("Files in %r: %s" % (cwd, files))

infile =  open(r'/tmp/CavAgent/interceptor_points.txt')
outfile = open(os.path.join(path, './test.py'))
temp = []
for x in infile:
    if x.startswith('#'):
        temp.append(x.lstrip('#').rstrip('\n'))

for line in outfile:
    count=0
    for y in temp:
        if y in line and 'from' not in line:
            f.write('#'+line)
            count = 1
    if count==0:
        f.write(line)


infile.close()
outfile.close()
f.close()
'''


def entry_point_refresh():
    #import os
    path = os.path.dirname(config.__file__)
    #print("Module path", path)
    #os.chdir(path)
    #print("Current working dir ==========================================",path)

    try:
        f = open(os.path.join(path, 'agent/internal/intercept_module.py'), "a+")
        #f.truncate(0)
    #cwd = os.getcwd()  # Get the current working directory (cwd)
    #files = os.listdir(cwd)  # Get all the files in that directory
    #print("Files in %r: %s" % (cwd, files))
    
        infile =  open(os.environ.get('ND_HOME') + '/CavAgent/interceptor_points.txt')
        outfile = open(os.path.join(path, './test.py'))
        temp = []
        for x in infile:
            if x.startswith('#'):
                temp.append(x.lstrip('#').rstrip('\n'))

        for line in outfile:
            count=0
            for y in temp:
                if y in line and 'from' not in line:
                    f.write('#'+line)
                    count = 1
            if count==0:
                f.write(line)


        infile.close()
        outfile.close()
        f.close()
    except:
        pass
    return 0




from pythonagent.agent.probes import add_hook, Instrumentation
import logging



_agent = None
#RUNNING_MODE = 0
from pythonagent.agent.internal.agent import Agent

def configure(environ=None):
    agent_config = config.parse_environ()
    if environ:
        agent_config.update(config.parse_environ(environ))

    config.merge(agent_config)
    configure_logging()

    return config.validate_config(agent_config)

all_module_names = list()

try:
    # import configparser
    # configParser = configparser.RawConfigParser()
    # configParser.read(r'config.txt')
    # path_for_insProfile = configParser.get('SO_PATH', 'path_profile')

    import os

    path = os.environ.get('ND_HOME')
    path = path + '/config/ndsettings.conf'
    myvars = {}
    with open(path) as myfile:
        for line in myfile:
            name, var = line.partition("=")[::2]
            myvars[name.strip()] = var
    try:
        path_for_insProfile = myvars.get('path_profile').strip('\n')
    except:
        path_for_insProfile = os.environ.get('ND_HOME') + "/CavAgent/instrumentationprofile.json"


    custom_methods_to_instrument_json = create_method_json(path_for_insProfile)["modules"]
    #all_module_names = list()

    for module in custom_methods_to_instrument_json:
        all_module_names.append(module["moduleName"])
except:
    pass

def bootstrap(agent=None):
    print("AGENT BOOTSTRAP")
    entry_point_refresh()
    import pythonagent.agent.internal.intercept_module
    reload(pythonagent.agent.internal.intercept_module)
    from pythonagent.agent.internal.intercept_module import BT
    BT_INTERCEPTORS = BT
    try:
        global _agent

        _agent = agent or Agent()
        hook = add_hook(_agent)
        #entry_point_refresh()
        #from pythonagent.agent.internal.intercept_module import BT as BT_INTERCEPTORS
        for module, patch in BT_INTERCEPTORS:
            hook.call_on_import(module, patch)

        for module in all_module_names:
            #print("MODULE NAME ", module)
            hook.call_on_import(module, Instrumentation.intercept_instrumented_method)

        _agent.module_interceptor = hook
        return _agent
    except:
        logging.getLogger('pythonagent.agent').exception('Error bootstrapping pythonagent; disabling.')
        return None

def get_agent_instance():
    if _agent is None:
        return bootstrap()
    else:
        return _agent


import time
from watchdog.observers.polling import PollingObserver
from watchdog.events import PatternMatchingEventHandler


def filewatchre(path, _agent):
    patterns = "*"
    ignore_patterns = ""
    ignore_directories = True
    case_sensitive = True
# Creating an event handler using watchdog.events
    my_event_handler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)

    def on_modified(event):
        _agent = get_agent_instance()
        bootstrap(_agent)

    my_event_handler.on_modified = on_modified

    path = path

    go_recursively = True
    my_observer = PollingObserver(timeout=5)
    my_observer.schedule(my_event_handler, path, recursive=go_recursively)

    my_observer.start()
filewatchre(os.environ.get('ND_HOME') + "/CavAgent/instrumentationprofile.json", _agent)
filewatchre(os.environ.get('ND_HOME') + "/CavAgent/interceptor_points.txt", _agent)




#import configparser
import errno
import os
#from agent.internal.proxy import Proxy
from main.pytrace import CommandInvocationError, CommandExecutionError
import sys

USAGE = "run <command> [args...]"

ABOUT = """Run a program with the agent enabled.

Use this command to instrument a Python application. If <command> is not the
path to an executable, the command is looked for in your PATH. The command may
be a Python interpreter or a supported server that embeds Python (like uwsgi
and gunicorn).

For example, if you normally run your application as below commands:

   1) python manage.py runserver --norelaod 10:20.0.117:6062
   2) gunicorn -w 4 -b unix:acme.sock acme.app:app

You can run the same application instrumented by pythonagent with below commands:
    
   1) cavagent run python manage.py runserver --norelaod 10:20.0.117:6062 
   2) cavagent run -- gunicorn -w 4 -b unix:acme.sock acme.app:app


The agent requires the TestRun to be running. If it
is not running, your application will not be start.

"""

OPTION = {
    'cli': 'run in CLI mode',
    
    # 'no-watchdog': 'disable the watchdog when auto-starting proxy',
    
    'config-file': {
        'short': 'c',
        'help': 'the config file to use',
        'value': True,
        'value_help': '<file>',
    },
    
    'app': {
        'short': 'a',
        'help': 'the name of the app',
        'value': True,
        'value_help': '<app>',
    },
    'tier': {
        'short': 't',
        'help': 'the name of the tier',
        'value': True,
        'value_help': '<tier>',
    },
    # 'node': {
    #     'short': 'n',
    #     'help': 'the name of the node',
    #     'value': True,
    #     'value_help': '<node>',
    # },

    # 'controller': {
    #     'short': 'h',
    #     'help': 'the host (and optionally port) of the controller',
    #     'value': True,
    #     'value_help': '<host>[:<port>]',
    # },

    # 'ssl': {
    #     'help': 'pass to use SSL with the controller',
    # },
    #
    # 'no-ssl': {
    #     'help': 'pass to disable SSL with the controller (the default)',
    # },

    'proxyMode': {
        'short': 'pmode',
        'help': 'proxyMode',
        'value': True,
        'value_help': '<proxyMode>  value - [0/1]',
    },

    'proxyConType': {
        'short': 'pctype',
        'help': 'please enter proxy con type ',
        'value': True,
        'value_help': '<proxyConType> value - [0/1]',
    },

    'proxyIP': {
        'short': 'pip',
        'help': 'please enter proxy IP host address',
        'value': True,
        'value_help': '<proxyConType>',
    },

    'proxyPort': {
        'short': 'pport',
        'help': 'please enter proxy port number ',
        'value': True,
        'value_help': '<proxyPort>',
    },

    # Internal / undocumented options

    # 'run-proxy-script': {
    #     'help': False,     # "Path to runProxy script"
    #     'value': True,
    # },
    #
    # 'proxy-args': {
    #     'help': False,  # "Command line arguments to pass to runProxy"
    #     'value': True,
    # },

}


def command(options, args):

    '''
    processname = 'TestRun'
    tmp = os.popen("nsu_show_netstorm").read()
    proccount = tmp.count(processname)

    if proccount > 0:
        print("TestRun is running...")
    else:
        print("ERROR!! TestRun is not running on this machine !!!\nKindly start the TestRun and try again...")
        sys.exit(1)
    '''
    import bootstrap
    pythonpath = [os.path.dirname(bootstrap.__file__), '.']

    if not args:
        raise CommandInvocationError('missing command: run <options> -- <command> [args...]')

    environ = os.environ

    if 'PYTHONPATH' in environ:
        pythonpath.append(environ['PYTHONPATH'])

    environ['PYTHONPATH'] = ':'.join(pythonpath)
    environ['LD_PRELOAD'] = "/usr/local/lib/libapr-1.so /usr/local/lib/libwebsockets.so /usr/local/lib/libcrypto.so /usr/local/lib/libssl.so"
    #print('ND_HOME set as ',environ.get('ND_HOME'))
    print("ND_HOME Environment variable :", environ.get('ND_HOME'))
    if environ.get('ND_HOME') is None:
        environ['ND_HOME'] = "/opt/cavisson/netdiagnostics"
        print("Taking Default ND_HOME path",environ.get('ND_HOME'))
    environ['CAV_RUNNING_MODE'] = '0'
    
    if 'config-file' in options:
        environ['CAV_CONFIG_FILE'] = options['config-file']
    if 'app' in options:
        environ['CAV_APP_NAME'] = options['app']
    if 'tier' in options:
        environ['CAV_TIER_NAME'] = options['tier']
    if 'node' in options:
        environ['CAV_NODE_NAME'] = options['node']
    if 'proxyMode' in options:
        environ['CAV_proxyMode'] = options['proxyMode']
    if 'proxyConType' in options:
        environ['CAV_proxyConType'] = options['proxyConType']
    if 'proxyIP' in options:
        environ['CAV_proxyIP'] = options['proxyIP']
    if 'proxyPort' in options:
        environ['CAV_proxyPort'] = options['proxyPort']
    environ['nd_init_done'] = '0'

    # if 'ssl' in options:
    #     environ['APPD_SSL_ENABLED'] = 'on'
    # elif 'no-ssl' in options:
    #     environ['APPD_SSL_ENABLED'] = 'off'

    # if 'controller' in options:
    #     value = options['controller']
    #
    #     if ':' in value:
    #         host, port = value.split(':', 1)
    #         environ['APPD_CONTROLLER_HOST'] = host
    #         environ['APPD_CONTROLLER_PORT'] = port
    #     else:
    #         environ['APPD_CONTROLLER_HOST'] = value

    # if 'use-manual-proxy' not in options:
    #     proxy_args = []
    #
    #     if 'no-watchdog' in options:
    #         proxy_args.append('--no-watchdog')
    #
    #     if 'run-proxy-script' in options:
    #         proxy_args.append('--run-proxy-script')
    #         proxy_args.append(options['run-proxy-script'])
    #
    #     if 'proxy-args' in options:
    #         proxy_args.extend(shlex.split(options['proxy-args']))

       # proxy.start(proxy_args)

    try:
        #proxy = Proxy.getInstance()
        #print("Proxy object created in run.py: ",proxy)

        #import bootstrap.sitecustomize
        #import agent.internal.agent as agent
        #agent.Agent().start()
        os.execvpe(args[0], args, environ)


    except OSError as exc:
        if exc.errno == errno.ENOENT:
            raise CommandExecutionError('%s: no such file or directory' % args[0])
        elif exc.errno == errno.EPERM:
            raise CommandExecutionError('%s: permission denied' % args[0])
        raise



import json

HAVOC_TYPES = [
    "InBoundService_Delay",
    "InBoundService_Failure",
    "OutBoundService_Delay",
    "OutBoundService_Failure",
    "MethodCall_Delay",
    "MethodCall_Failure",
    "Custom_Memory_Leak"
]

# EXPIRE_DURATION_MS = 300000  # 5 minutes
# EXPIRE_DURATION_MS = 3000  # 3 seconds
EXPIRE_DURATION_MS = 1000  # 1 seconds


SIZE_OF_OBJECTS_BYTES = 524288000  # 500 MB

# Inbound Service Delay
IB_SERV_DELAY_DICT = {
   "enable": 0,
   "havocType": 1,
   "bTName": "",
   "uRL": "/prod/path/to/resource",
   "methodFQM": "",
   "delayInSec": 10,
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
IB_SERV_DELAY_JSON = json.dumps(IB_SERV_DELAY_DICT)

# Inbound Service Failure
IB_SERV_FAIL_DICT = {
   "enable": 0,
   "havocType": 2,
   "bTName": "",
   "uRL": "/prod/path/to/resource",
   "methodFQM": "",
   "delayInSec": 10,
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
IB_SERV_FAIL_JSON = json.dumps(IB_SERV_FAIL_DICT)

# Outbound Service Delay
OB_SERV_DELAY_DICT = {
   "enable": 0,
   "havocType": 3,
   "bTName": "",
   "uRL": "",
   "methodFQM": "",
   "delayInSec": 10,
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "sample_hostname",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
OB_SERV_DELAY_JSON = json.dumps(OB_SERV_DELAY_DICT)

# Outbound Service Failure
OB_SERV_FAIL_DICT = {
   "enable": 0,
   "havocType": 4,
   "bTName": "",
   "uRL": "",
   "methodFQM": "",
   "delayInSec": 0,
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "sample_hostname",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
OB_SERV_FAIL_JSON = json.dumps(OB_SERV_FAIL_DICT)

# Method Call Delay
METH_CALL_DELAY_DICT = {
   "enable": 0,
   "havocType": 5,
   "bTName": "",
   "uRL": "",
   "methodFQM": "/var/task.lambda_handler",
   "delayInSec": 10,
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
METH_CALL_DELAY_JSON = json.dumps(METH_CALL_DELAY_DICT)

# Method Invocation Failure
METH_INV_FAIL_DICT = {
   "enable": 0,
   "havocType": 6,
   "bTName": "",
   "uRL": "",
   "methodFQM": "/var/task.lambda_handler",
   "delayInSec": "",
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "",
   "backendName": "",
   "leaksINMB": "",
   "memoryLeaksInSec": ""
}
METH_INV_FAIL_JSON = json.dumps(METH_INV_FAIL_DICT)

# Custom Memory Leak
CUSTOM_MEMORY_LEAK_DICT = {
   "enable": 0,
   "havocType": 7,
   "bTName": "",
   "uRL": "",
   "methodFQM": "/var/task.lambda_handler",
   "delayInSec": "",
   "totalDurationInSec": 20,
   "threshOld": {"count": 0, "percentage": 0, "currentcount": 0},
   "protocol": "",
   "hostname": "",
   "backendName": "",
   "leaksINMB": 500,
   "memoryLeaksInSec": 5
}
CUSTOM_MEMORY_LEAK_JSON = json.dumps(CUSTOM_MEMORY_LEAK_DICT)


HAVOC_PROFILES_JSON = [IB_SERV_DELAY_JSON, IB_SERV_FAIL_JSON,
                       OB_SERV_DELAY_JSON, OB_SERV_FAIL_JSON,
                       METH_CALL_DELAY_JSON, METH_INV_FAIL_JSON,
                       CUSTOM_MEMORY_LEAK_JSON]



import json
from collections import namedtuple
import time
import logging
from pythonagent.agent.internal import udp
from pythonagent.agent.probes.havoc.havoc_constants import HAVOC_TYPES, EXPIRE_DURATION_MS, SIZE_OF_OBJECTS_BYTES, HAVOC_PROFILES_JSON
# from pythonagent.agent.probes.havoc.custom_memory_stress import apply_memory_leak
from pythonagent.utils import get_current_timestamp_in_ms

havoc_types = {
    1: "InBoundService_Delay",
    2: "InBoundService_Failure",
    3: "OutBoundService_Delay",
    4: "OutBoundService_Failure",
    5: "MethodCall_Delay",
    6: "MethodCall_Failure",
    7: "Custom_Memory_Leak"
}


class NDHavocException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class NDNetHavocRequest(object):
    def __init__(self, havoc_conf, header_dict):

        self.logger = logging.getLogger('pythonagent.agent')
        self.startTime = get_current_timestamp_in_ms()

        self.header_dict = header_dict
        self.nhmid = header_dict["NHMID"]

        self.havocType = havoc_conf.havocType
        self.totalDurationInSec = havoc_conf.totalDurationInSec

        if hasattr(havoc_conf, 'delayInSec'):
            self.delayInSec = havoc_conf.delayInSec / 1000  # UI is sending in milliseconds

        if self.havocType in [1, 2]:
            self.bTNameMode = havoc_conf.bTNameMode
            if hasattr(havoc_conf, 'uRL'):
                self.uRL = havoc_conf.uRL

        if self.havocType in [3, 4]:
            self.backendNameMod = havoc_conf.backendNameMod
            if hasattr(havoc_conf, 'hostname'):
                self.hostname = havoc_conf.hostname

        if self.havocType in [5, 6]:
            self.methodMod = havoc_conf.methodMod
            if hasattr(havoc_conf, 'methodFQM'):
                self.methodFQM = havoc_conf.methodFQM

        if self.havocType in [7]:
            self.leaksINMB = havoc_conf.leaksINMB
            self.memoryLeaksInSec = havoc_conf.memoryLeaksInSec

        try:
            self.enable = havoc_conf.enable
            self.bTName = havoc_conf.bTName
            self.isHavocEnable = False
            self.threshOld = havoc_conf.threshOld
            self.protocol = havoc_conf.protocol
            self.backendName = havoc_conf.backendName

        except Exception as e:
            self.logger.warning("Optional parameters not found {}".format(e))


class NDNetHavocMonitor(object):

    __instance = None

    @staticmethod
    def get_instance():
        if NDNetHavocMonitor.__instance == None:
            NDNetHavocMonitor()
        return NDNetHavocMonitor.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if NDNetHavocMonitor.__instance != None:
            raise Exception("This class is a singleton! Use get_instance()")
        else:
            NDNetHavocMonitor.__instance = self

            self.enableNetHavoc = False

            self.netHavocConfigMap = {
                "InBoundService_Delay": [],
                "InBoundService_Failure": [],
                "OutBoundService_Delay": [],
                "OutBoundService_Failure": [],
                "MethodCall_Delay": [],
                "MethodCall_Failure": [],
                "Custom_Memory_Leak": []
            }

            self.logger = logging.getLogger('pythonagent.agent')

    def set_enable_net_havoc(self, enable_logs_net_havoc):
        try:
            if enable_logs_net_havoc == 1:
                self.enableNetHavoc = True
            else:
                self.enableNetHavoc = False
        except Exception as e:
            raise e

    def validate_config(self, havoc_conf):
        self.logger.info("Validating config")
        is_valid = False

        # Validate enable
        if not hasattr(havoc_conf, 'enable'):
            return_message = "enable not found"
            return is_valid, return_message

        else:
            if havoc_conf.enable not in [0, 1]:
                return_message = "invalid enable not found"
                return is_valid, return_message
            else:
                logging.debug("enable validated")

        # Validate havocType
        if not hasattr(havoc_conf, 'havocType'):
            return_message = "havocType not found"
            return is_valid, return_message
        else:
            if havoc_conf.havocType not in [1, 2, 3, 4, 5, 6, 7, 8]:
                return_message = "invalid havocType"
                return is_valid, return_message
            else:
                havoc_type = havoc_conf.havocType
                logging.debug("havocType validated")

        # Validate totalDurationInSec
        if not hasattr(havoc_conf, 'totalDurationInSec'):
            return_message = "totalDurationInSec not found"
            return is_valid, return_message
        else:
            if not isinstance(havoc_conf.totalDurationInSec, int):
                return_message = "invalid totalDurationInSec"
                return is_valid, return_message
            else:
                logging.debug("totalDurationInSec validated")

        # Validate threshOld
        if not hasattr(havoc_conf, 'threshOld'):
            return_message = "threshOld not found"
            return is_valid, return_message
        else:
            if not hasattr(havoc_conf.threshOld, 'percentage'):
                return_message = "invalid threshOld"
                return is_valid, return_message
            else:
                self.logger.debug("threshOld validated")

        # Validate uRL for Havoc Type 1 and 2
        if havoc_type in [1, 2]:
            self.logger.debug("bTNameMode: {}".format(havoc_conf.bTNameMode))
            if havoc_conf.bTNameMode == 0:
                pass
            elif havoc_conf.bTNameMode == 3:
                if not hasattr(havoc_conf, 'uRL'):
                    return_message = "uRL not found for inbound service"
                    return is_valid, return_message
                else:
                    if len(havoc_conf.uRL) == 0:
                        return_message = "invalid uRL  for inbound service"
                        return is_valid, return_message
                    else:
                        self.logger.debug("uRL validated")
            else:
                return_message = "bTNameMode 1 and 2 not supported"
                return is_valid, return_message
        else:
            self.logger.debug("uRL not needed")

        # Validate hostname and protocol for Havoc Type 3 and 4
        if havoc_type in [3, 4]:
            if havoc_conf.backendNameMod == 0:
                pass
            elif havoc_conf.backendNameMod == 3:
                if not hasattr(havoc_conf, 'hostname'):
                    return_message = "hostname not found for outbound service"
                    return is_valid, return_message
                else:
                    if len(havoc_conf.hostname) == 0:
                        return_message = "invalid hostname for outbound service"
                        return is_valid, return_message
                    else:
                        self.logger.debug("hostname validated")

                    if not hasattr(havoc_conf, 'protocol'):
                        if havoc_conf.protocol not in ["Http", "Non Http"]:
                            return_message = "invalid protocol for outbound service"
                            return is_valid, return_message
                        else:
                            self.logger.debug("protocol validated")
                    else:
                        self.logger.debug("protocol not found")
            else:
                return_message = "backendNameMod 1, 2, and 4 not supported"
                return is_valid, return_message

        else:
            self.logger.debug("hostname not needed")

        # Validate methodFQM
        if havoc_type in [5, 6]:
            if not hasattr(havoc_conf, 'methodFQM'):
                return_message = "methodFQM not found for method"
                return is_valid, return_message
            else:
                if len(havoc_conf.methodFQM) == 0:
                    return_message = "invalid methodFQM for method"

                    return is_valid, return_message
                else:
                    self.logger.debug("methodFQM validated")
        else:
            self.logger.debug("methodFQM not needed")

        if havoc_type in [1, 3, 5]:
            if not hasattr(havoc_conf, 'delayInSec'):
                return_message = "delayInSec not found for delay"
                return is_valid, return_message
            else:
                if not isinstance(havoc_conf.delayInSec, int):
                    return_message = "invalid delayInSec for delay"
                    return is_valid, return_message
                else:
                    self.logger.debug("delayInSec validated")
        else:
            self.logger.debug("delayInSec not needed")

        if havoc_type in [7]:
            if not hasattr(havoc_conf, "leaksINMB"):
                return_message = "leaksINMB not found for custom_memory_leak"
                return is_valid, return_message
            else:
                if not hasattr(havoc_conf, "memoryLeaksInSec"):
                    return_message = "memoryLeaksInSec not found for custom_memory_leak"
                    return is_valid, return_message
                else:
                    self.logger.debug("memoryLeaksInSec validated")
        else:
            self.logger.debug("leaksINMB, memoryLeaksInSec not needed")

        is_valid = True
        return_message = "Config Validated"
        return is_valid, return_message

    def parse_nethavoc_config(self, json_conf, header_dict):
        self.logger.info("Parsing config")

        try:
            havoc_conf = json.loads(json_conf, object_hook=lambda d: namedtuple('X', d.keys())(*d.values()))
            havoc_type = havoc_conf.havocType
            self.logger.info("Havoc type: {}".format(havoc_type))
            self.logger.debug("header_dict: {}".format(header_dict))
            nhmid = header_dict["NHMID"]
            self.logger.debug("nhmid: {}".format(nhmid))
            self.logger.info("checking if havoc already exists")

            havoc_type_full_str = havoc_types[havoc_type]
            config_list = self.netHavocConfigMap[havoc_type_full_str]
            self.logger.info("config_list: {}".format(config_list))

            if config_list:
                for config in config_list:
                    if nhmid == config.nhmid:
                        self.logger.info("havoc already exists skipping")
                        return

            is_valid, ret_message = self.validate_config(havoc_conf)

            if is_valid:
                self.logger.info(ret_message)
                req_obj = NDNetHavocRequest(havoc_conf, header_dict)
                self.logger.debug("Request Object: ".format(req_obj))
                havoc_type_full_str = havoc_types[havoc_type]
                self.netHavocConfigMap[havoc_type_full_str].append(req_obj)

            else:
                self.logger.info("Invalid config: {}".format(ret_message))

        except Exception as e:
            self.logger.error("havoc parsing exception: {}".format(e))
            raise e

    def common_delay_in_response(self, total_duration_in_sec):
        self.logger.info("Applying delay duration(sec): {}".format(total_duration_in_sec))
        try:
            time.sleep(total_duration_in_sec)
        except Exception as e:
            raise e

    def common_failure(self, exception_message):
        self.logger.info("Applying failure message: {}".format(exception_message))


        raise NDHavocException(exception_message)

    @staticmethod
    def create_dummy_object(obj_size):
        int_list_size = obj_size // 4  # Size of int = 4
        obj = []
        for i in range(int_list_size):
            obj.append(i)
        return obj

    def memory_leak(self, leaksINMB, memoryLeaksInSec):

        total_leak_bytes = leaksINMB * 1024 * 1024
        num_of_obj = total_leak_bytes // SIZE_OF_OBJECTS_BYTES
        time_for_each_obj = memoryLeaksInSec // num_of_obj

        obj_created = 0
        while obj_created < num_of_obj:
            self.create_dummy_object(SIZE_OF_OBJECTS_BYTES)
            time.sleep(time_for_each_obj)
            obj_created += 0

    def apply_inbound_service_delay(self, req_url):
        self.logger.info("Trying to apply inbound service delay")
        config_list = self.netHavocConfigMap["InBoundService_Delay"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                delay_in_second = config.delayInSec

                if config.bTNameMode == 0:
                    self.logger.debug("bTNameMode = 0, proceeding")
                    self.common_delay_in_response(delay_in_second)
                else:
                    url = config.uRL
                    self.logger.debug("url: {} in req_url: {}".format(url, req_url))
                    if url in req_url:
                        self.logger.debug("Config matched for Request URL {}".format(req_url))
                        self.common_delay_in_response(delay_in_second)
                    else:
                        self.logger.debug("Config not matched for Request URL {}".format(req_url))

        else:
            self.logger.debug("config not set for apply_inbound_service_delay")

    def apply_inbound_service_failure(self, req_url):
        self.logger.info("Trying to apply inbound service failure")
        config_list = self.netHavocConfigMap["InBoundService_Failure"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                if config.bTNameMode == 0:
                    self.logger.debug("bTNameMode = 0, proceeding")
                    self.common_failure("BT:ALL")
                else:
                    url = config.uRL
                    self.logger.debug("url: {} in req_url: {}".format(url, req_url))
                    if url in req_url:
                        self.logger.debug("Config matched for Request URL {}".format(req_url))
                        self.common_failure(url)
                    else:
                        self.logger.debug("Config not matched for Request URL {}".format(req_url))
        else:
            self.logger.debug("config not set for apply_inbound_service_failure")
            # raise Exception("config not set for apply_inbound_service_failure")

    def apply_outbound_service_delay(self, req_hostname):
        self.logger.info("Trying to apply outbound service delay")
        config_list = self.netHavocConfigMap["OutBoundService_Delay"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                delay_in_second = config.delayInSec
                if config.backendNameMod == 0:
                    self.logger.debug("backendNameMod = 0, proceeding")
                    self.common_delay_in_response(delay_in_second)
                else:
                    hostname = config.hostname
                    self.logger.debug("hostname: {} in req_hostname: {}".format(hostname, req_hostname))

                    if hostname in req_hostname:
                        self.logger.debug("Config matched for Request Hostname {}".format(req_hostname))
                        self.common_delay_in_response(delay_in_second)
                    else:
                        self.logger.debug("Config not matched for Request Hostname {}".format(req_hostname))
        else:
            self.logger.debug("config not set for apply_outbound_service_delay")

    def apply_outbound_service_failure(self, req_hostname):
        self.logger.info("Trying to apply outbound service failure")
        config_list = self.netHavocConfigMap["OutBoundService_Failure"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                if config.backendNameMod == 0:
                    self.logger.debug("backendNameMod = 0, proceeding")
                    self.common_failure("Integration Point: ALL")
                else:
                    hostname = config.hostname
                    self.logger.debug("hostname: {} in req_hostname: {}".format(hostname, req_hostname))

                    if hostname in req_hostname:
                        self.logger.debug("Config matched for Request Hostname {}".format(req_hostname))
                        self.common_failure(hostname)
                    else:
                        self.logger.debug("Config not matched for Request Hostname {}".format(req_hostname))
        else:
            self.logger.debug("config not set for apply_outbound_service_failure")

    def apply_method_call_delay(self, req_method_fqm):
        self.logger.info("Trying to apply method call delay")
        config_list = self.netHavocConfigMap["MethodCall_Delay"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                delay_in_second = config.delayInSec
                method_fqm = config.methodFQM
                self.logger.debug("method_fqm: {} in req_method_fqm: {}".format(method_fqm, req_method_fqm))

                if method_fqm in req_method_fqm:
                    self.logger.debug("Config matched for Request FQM {}".format(req_method_fqm))
                    self.common_delay_in_response(delay_in_second)
                else:
                    self.logger.debug("Config not matched for Request FQM {}".format(req_method_fqm))
        else:
            self.logger.debug("config not set for apply_method_call_delay")

    def apply_method_invocation_failure(self, req_method_fqm):
        self.logger.info("Trying to apply method invocation failure")
        config_list = self.netHavocConfigMap["MethodCall_Failure"]

        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                method_fqm = config.methodFQM
                self.logger.debug("method_fqm: {} in req_method_fqm: {}".format(method_fqm, req_method_fqm))

                if method_fqm in req_method_fqm:
                    self.logger.debug("Config matched for Request FQM {}".format(req_method_fqm))
                    self.common_failure(method_fqm)
                else:
                    self.logger.debug("Config not matched for Request FQM {}".format(req_method_fqm))
        else:
            self.logger.debug("config not set for apply_method_call_failure")

    def increase_memory_usage(self):
        self.logger.info("Trying to apply memory leak")
        config_list = self.netHavocConfigMap["Custom_Memory_Leak"]
        if config_list:
            for config in config_list:
                self.logger.debug("Trying to apply havoc: {}".format(config.nhmid))
                leaksINMB = config.leaksINMB
                totalDurationInSec = config.totalDurationInSec
                memoryLeaksInSec = config.memoryLeaksInSec
                apply_memory_leak(leaksINMB, totalDurationInSec, memoryLeaksInSec)

        else:
            self.logger.debug("config not set for apply_method_call_failure")

    def increase_cpu_utilization(self):
        pass

    def refresh_configs(self, agent_obj):
        self.logger.info("Refreshing Configs")
        current_time_ms = get_current_timestamp_in_ms()

        for havoc_type in HAVOC_TYPES:
            self.logger.info("Checking havoc type: {}".format(havoc_type))

            config_list = self.netHavocConfigMap[havoc_type]
            if config_list:
                self.logger.info("Config found for havoc type: {}".format(havoc_type))
                for config in list(config_list):
                    self.logger.debug("Checking Config {} for type {}".format(id(config), config.havocType))

                    total_duration_in_ms = config.totalDurationInSec * 1000
                    self.logger.debug("total_duration_in_ms: {}".format(total_duration_in_ms))

                    start_time_ms = config.startTime
                    duration = current_time_ms - start_time_ms
                    self.logger.debug("duration: {}".format(duration))

                    # if duration >= EXPIRE_DURATION_MS:
                    if duration >= total_duration_in_ms:
                        self.logger.debug("Config {} for type {} expired, removing".format(id(config), config.havocType))
                        config_list.remove(config)

                        header_str = "NetDiagnosticMessage2.0;"
                        for key, value in config.header_dict.items():
                            header_str += key + ":" + value + ";"

                        udp.havoc_message(agent_obj, header_str)

                    else:
                        self.logger.debug("Config {} for type {} not expired, continuing".format(id(config), config.havocType))
            else:
                self.logger.debug("config not found for havoc type {}".format(havoc_type))



import os
import sys
import cProfile
import threading
import pstats
from collections import Counter
from pyclbr import readmodule

class Aisession():
    def __init__(self, istart, starttime, duration, sess):
        self.duration = duration
        self.istart = istart
        self.starttime = starttime
        self.sess = sess
        print("SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSEEEEEEEEEEEEEEEEEEEEEEEEEEEESSSSSSSSSSSSSSSSSSSSSSSSSSSSSS", sess)
        self.prof = cProfile.Profile()
        os.environ['aisess_status'] = 'True' 

    def enableprof(self):
        self.prof.enable()
        #os.environ['aisess_status'] = 'True'
        #os.environ['Aisession_flag'] = 'True'

    def endsession(self):
        self.prof.disable()
        self.printst()
        os.environ['aisess_status'] = 'False'
        os.environ['Aisession_flag'] = 'False'
        print("@!#@$#%$^&*^&%^$%#$@####$#%$^% IN endsession")
        self.prof = None
        
    """def disableprof(self):
        timer = threading.Timer(self.time,self.endsession)
        timer.start()"""
    def disableprof(self):
        #if self.istart == 0:
        #   self.endsession()
        timer = threading.Timer(self.duration,self.endsession)
        timer.start()
        print("exiting endsession")


    """def stopaisession(self, duration):
        self.disableprof(duration)
        print(" IN stopaisession")
        self.printst()
        self.aiobj = None"""


    def printst(self):
        path =  "/opt/cavisson/netdiagnostics/logs/"
        filename = str(self.sess.decode('ascii'))+"_AI.txt"
        file_to_open = os.path.join(path, filename)
        filenameai = str(self.sess.decode('ascii'))+".txt"
        aifile_to_open = os.path.join(path, filenameai)
        xmlfilename = str(self.sess.decode('ascii'))+".xml"
        xmlfile_to_open = os.path.join(path, xmlfilename)
        xmlfile = open(xmlfile_to_open, 'w')
        f = open(file_to_open, 'w')
        f2 = open(aifile_to_open, 'w')
        #f = open(f'/opt/cavisson/netdiagnostics/logs/{self.sess}.txt', 'w')
        f.writelines("************************************  AUTO-INSTRUMENTATION  **************************************** \n")
        f2.writelines("************************************  AUTO-INSTRUMENTATION  **************************************** \n")
        f.writelines("# <FQM>| <I:(instrumented) / C :(consisdered) / R :(remove) >| < Total Count > | < Total Duration > | < Added TimeStamp > | < Remove TimeStamp > \n")
        f.writelines(f"#Start_Time {self.starttime}\n")
        #f.writelines(f"#End_Time {}")
        f.writelines(f"#Total Duration {self.duration}\n")
        p = pstats.Stats(self.prof)
        client_app_root = os.path.dirname(sys.argv[0])
        print("application:-",client_app_root)
        methodcountdict = {}
        for k,v in p.stats.items():
            #print(k[0])
            if client_app_root in k[0]:
                modulek001 = k[0]
                print("$$$$$$$$$$$$$$$$$$", modulek001)
                module = k[0]
                print("module:-", k[0])
                module_path = os.path.split(os.path.abspath(module))[0]
                modulefullname = os.path.split(os.path.abspath(module))[1]
                module_name = modulefullname.split('.')[0]
                methodname = k[-1]
                modulemethod = module_path+'*'+module_name+'*'+methodname
                total_call = str(v[1])
                total_duration =str(v[3])
                print('modulemethod:-', modulemethod)
                print('module path:-', module_path)
                print('module name:-', module_name)
                print('method name:- ',k[-1])
                print('Total call:- ',v[1])
                print('Response time of method:-',v[3])
                #methodcountdict = {}
                methodcountdict[modulemethod] = total_call
                mod = readmodule(module_name, path=[module_path])
                li = []
                for k, v in mod.items():
                    li.append(k)
                    methods = v.methods.items()
                    lk = []
                    for method, lineno in methods:
                        lk.append(method)
                    li.append(lk)
                count=0
                classname =''
                for i in li:
                    count +=1
                    if type(i)==str:
                        pass
                    else:
                        for j in i:
                            if j == methodname:
                                classname = li[count-2]
                               #print("classname:-",li[count-2])
                                varfullname = module_name+"."+li[count-2]+"."+methodname
                               #print("))))))))))))*((((((((((((",a)
                            else:
                                pass
                if len(classname)>0:
                    print(varfullname)
                else:
                    varfullname = module_name+"."+methodname
                    print(varfullname)
                outputform = varfullname+"|"+"I"+"|"+total_call+"|"+total_duration+"|"+"0"+"|"+"0"
                print("************************************XWWDWSASSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS", outputform)
               # f.writelines("************************************  AUTO-INSTRUMENTATION  **************************************** \n")
                f.writelines(f"{outputform}  \n")
                if (float(total_duration)/int(total_call)) >=0.7:
                   outputformai = varfullname+"|"+"I"+"|"+total_call+"|"+total_duration+"|"+"0"+"|"+"0"
                   f2.writelines(f"{outputformai}  \n")
        f.close()    
        f2.close()
        xmlfile.close()
        print(methodcountdict)            

"""def aisession():
    ai = WSGIInterceptor()
    ai.application_callable = agentprofiler.application_callable_pro

#aisession()

def ta():
    wsgi.WSGIInterceptor.application_callable = agentprofiler.application_callable

#if flag ==True:
#timer = threading.Timer(300, ta)
#timer.start()
def enablendisablesess():
    aisession()
    timer = threading.Timer(300, ta)
    timer.start()"""    




"""Interceptors for dealing with DB-API 2.0 compatible libraries.

"""

from __future__ import unicode_literals
import abc
import re
#import agent
#from agent.internal.proxy import *
from pythonagent.lang import str
#from pythonagentynamics.agent.core.registries import MissingConfigException
#from pythonagentynamics.agent.models.exitcalls import EXIT_DB, EXIT_SUBTYPE_DB
#from .. import HOST_PROPERTY_MAX_LEN, DB_NAME_PROPERTY_MAX_LEN, VENDOR_PROPERTY_MAX_LEN
from ..base import ExitCallInterceptor

EXIT_DB = 1
EXIT_SUBTYPE_DB = 'DB'

HOST_PROPERTY_MAX_LEN = 50
DB_NAME_PROPERTY_MAX_LEN = 50
VENDOR_PROPERTY_MAX_LEN = 50

class MissingConfigException(Exception):
    pass


#def get_db_backend(agent, host, port, dbname, vendor):
    # if agent.backend_registry is None:
    #     raise MissingConfigException

#    naming_format_string = '{HOST}:{PORT} - {DATABASE} - {VENDOR} - {VERSION}'
#    host = host[:HOST_PROPERTY_MAX_LEN]
#    dbname = dbname[:DB_NAME_PROPERTY_MAX_LEN]
#    vendor = vendor[:VENDOR_PROPERTY_MAX_LEN]

#    backend_properties = {
#        'VENDOR': vendor,
#        'HOST': host,
#        'PORT': str(port),
#        'DATABASE': dbname,
#        'VERSION': 'unknown',
#    }

    #return agent.backend_registry.get_backend(EXIT_DB, EXIT_SUBTYPE_DB, backend_properties, naming_format_string)
#    return None


class DbAPICursorInterceptor(ExitCallInterceptor):
    
    #print("going to get proxy instance (DbAPICursor Interceptor)")
    #proxy = Proxy.getInstance()
    #print("proxy handle = ",proxy )

    def db_call_end(self, bt,exit_call):
        #self.agent.logger.info('Modulename: DbAPICursorInterceptor class inside db_call_end function') 
        # execute_result is intentionally not used here.  Its purpose is to
        # allow subclasses to override this method and use the value.
        super(DbAPICursorInterceptor, self).db_call_end(bt,exit_call)

    def _execute(self, execute, cursor, operation, *args, **kwargs):
        with self.log_exceptions():
            print("operation in  db query in dbapi.py\n\n", operation)
            backend = True
            exit_call = None
           # host = None
            #self.agent.logger.info("inside _execute funciton")
            #self.agent.logger.info('Modulename: DbAPICursorInterceptor class inside _execute function')
            self.agent.logger.info("Modulename: DbAPICursorInterceptor class || execute value is : {0}".format(execute))
            self.agent.logger.info("Modulename: DbAPICursorInterceptor class || operation value is :{0}".format(str(operation)))
            #self.agent.logger.info("execute: {}",execute,"cursor: {}",cursor,"operation:{} ",operation)
            #print("cursor host: ", cursor.host)
           
            #bt = self.agent.bt
            bt = self.bt
            if bt:
                #print("bt value is ------------------",bt)
                connection = self.get_connection(cursor)
                #print("connection",str(connection))
               # hosts = list()
               # regex1 =r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'
               # regex2 = "localhost"
               # regexList = [regex1, regex2]
                #host = None
               # print("regexList---------------------->",regexList)
               # for regex in regexList:
                #    h = re.search(regex,str(connection))
                 #   hosts.append(h)           
                   # print("ho host = re.search(regex,str(connection))up())   
               # host =  next(h for h in hosts if h is not None)
               # host = host.group() 
               # print("host is------------------------>",host)
                try:
                    #backend = connection._pythonagent_backend
                    #self.agent.logger.info("connection.host:", connection.dsn)
                   # print("inside exute function try block")
                    self.agent.logger.info("Modulename: DbAPICursorInterceptor class || connection.host value is :{0}".format(connection.dsn))
                    from pythonagent.agent.probes.sql.psycopg2 import parse_postgresql_kv_dsn
                    host, port, dbname = parse_postgresql_kv_dsn(connection)
                except AttributeError:
                    #print("types of cursor:",type(cursor))
                     try:
                         host = connection.host
                     except:
                         hosts = list()
                         regex1 =r'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
                         regex2 =r'localhost'
                         regexList = [regex1, regex2]
                         for regex in regexList:
                             h= re.search(regex,str(connection))
                             hosts.append(h)
                            # if host is not None:
                               # host = host.group()
                         #print("host is------------------>",hosts)
                         host =  next(h for h in hosts if h is not None)
                         host = host.group()
                         #print("host is------------------>",host)

                    # If we weren't able to create a backend when the
                    # connection object was created, we can try again now.
                    # try:
                    #     backend = get_db_backend(self.agent, *connection._pythonagent_backend_properties)
                    #     connection._pythonagent_backend = backend
                    # except MissingConfigException:
                    #     # Still no config; try again next time!
                    #     pass
                #self.agent.logger.info("connection: ",connection,"backend: ",backend)
                self.agent.logger.info("Modulename: DbAPICursorInterceptor class || connection value is :{0}".format(connection))
                self.agent.logger.info("Modulename: DbAPICursorInterceptor class || backend is :{0}".format(backend))
                if backend and not hasattr(cursor, '_pythonagent_exit_call'):
                    #print("inside inner if function : ---------------------------------")
                    # exit_call = self.proxy.db_call_begin(bt,host,operation)
                    exit_call = self.db_call_begin(bt,host,operation)
                    cursor._pythonagent_exit_call = exit_call
                
             
        result = execute(cursor, operation, pythonagent_exit_call=exit_call, *args, **kwargs)
        if str(operation) == 'SELECT * FROM products':
            import time
            time.sleep(2)
        else:
            pass
        # Normally it's fine to just try to end the exit call even if we
        # aren't sure it actually started.  However for cases where we have
        # several `execute` calls nested inside an `executemany` call, we do
        # not want to end the exit call until the outer call is over.
        if exit_call:
            #self.proxy.db_call_end(bt,exit_call)
            self.db_call_end(bt,exit_call)
            try:
                delattr(cursor, '_pythonagent_exit_call')
            except AttributeError:
                pass
        return result

    @abc.abstractmethod
    def get_connection(self, cursor):
        """Return the `Connection` object used to create this `Cursor`.

        """
        pass


class DbAPIConnectionInterceptor(ExitCallInterceptor):
    #print("going to get proxy instance (DbAPIDbAPIConnection Interceptor)")
    #proxy = Proxy.getInstance()
    #print("proxy handle = ",proxy )
    def __init__(self, agent, cls, cursor_interceptor):
        super(DbAPIConnectionInterceptor, self).__init__(agent, cls)
        self.cursor_interceptor = cursor_interceptor

    def attach(self, connect_func):
        super(DbAPIConnectionInterceptor, self).attach(connect_func, patched_method_name='_connect')
        super(DbAPIConnectionInterceptor, self).attach('cursor')

    def db_call_end(self, bt,exit_call):
        # maybe_connection is intentionally not used here.  Its purpose is to
        # allow subclasses to override this method and use the value.
        super(DbAPIConnectionInterceptor, self).db_call_end(bt,exit_call)

    def _connect(self, connect, connection, *args, **kwargs):
        bt = self.bt
        with self.log_exceptions():
            backend = True
            exit_call = None

            backend_properties = self.get_backend_properties(connection, *args, **kwargs)
            #print('Type of backend property: ',backend_properties)
            self.agent.logger.info('backend_properties value'.format(backend_properties))
            try:
                #backend = get_db_backend(self.agent, *backend_properties)
                #print("backend------->",backend)
                connection._pythonagent_backend = backend
            except MissingConfigException:
                # If the agent has just started and doesn't have config yet,
                # we cannot create and store a backend on the connection object.
                # We store the backend properties instead so that the cursor
                # interceptor can try again later.
                connection._pythonagent_backend_properties = backend_properties
            #from agent.probes.sql.psycopg2 import parse_postgresql_kv_dsn
            #host, port, dbname = parse_postgresql_kv_dsn(connection.dsn)
            #print('connection ', connection)
            #print('connect ', connect)
            #bt = self.agent.bt
           # if bt and backend:
            if bt :
                self.agent.logger.info('Start db call for DBAPI connection interceptor')
                #exit_call = self.proxy.db_call_begin(bt,backend_properties[0],'connect')
               # print("connect--------------------------------------------",connect)
                #print("connection--------------------------------------------",connection)
                exit_call = self.db_call_begin(bt,backend_properties[0],'connect')

        # The reason this is called `maybe_connection` is that different
        # interceptors will return different values here.  `connect` could
        # be an __init__ method on a class, returning None, it could be a
        # function which returns an initialized connection object, or even a
        # method returning an async future object.  The only reason we capture
        # this value and pass it to `end_exit_call` is to allow subclasses to
        # override the behaviour of `end_exit_call` (for example in the async
        # case, to only end the exit call when the future is completed).
        maybe_connection = connect(connection, pythonagent_exit_call=exit_call, *args, **kwargs)
        #self.end_exit_call(exit_call, maybe_connection)
        self.agent.logger.info('end db call for DBAPI connection interceptor')
        #self.proxy.db_call_end(bt,exit_call)
        self.db_call_end(bt,exit_call)
        return maybe_connection

    def _cursor(self, cursor, connection, *args, **kwargs):
        cursor_instance = cursor(connection, pythonagent_exit_call=None, *args, **kwargs)
        with self.log_exceptions():
            self.cursor_interceptor(self.agent, type(cursor_instance)).attach(['execute', 'executemany'],
                                                                              patched_method_name='_execute')
        return cursor_instance

    @abc.abstractmethod
    def get_backend_properties(self, connection, *args, **kwargs):
        """Return a tuple of (host, port, database, vendor) for this connection.

        The parameters passed to this function are the same as those passed to
        the intercepted `connect` function.

        WARNING: `connection` may not have been initialized when this
        function is called; be careful.

        """
        pass



from __future__ import unicode_literals
from ..base import ExitCallInterceptor
from pythonagent.lang import str


def intercept_elastic(agent, mod):
    class ExitCallListener(mod.monitoring.CommandListener):
        backend_name_format_string = '{HOST}:{PORT} - {DATABASE}'
    
        def __init__(self):
            self.interceptor = ExitCallInterceptor(agent, None)
            self.exit_call_map = {}
            
    
        def get_backend(self, connection_id):
            #agent.logger.debug('Modulenameintercept_pymongo class inside get_backend function')
            print("connection iddddddddddddddddddddddddddddddddd::::::::::::::::::", connection_id)
            host, port = connection_id
            backend_properties = {
                'HOST': host,
                'PORT': str(port)
            }
            return backend_properties
        
        def started(self, event):
            agent.logger.info('Modulenameintercept_elastic  class inside started  function  is ')
            with self.interceptor.log_exceptions():
                agent.logger.info('Modulenameintercept_elastic  class inside started  function  is ')
            with self.interceptor.log_exceptions():
                agent.logger.info("inside _execute funciton")
                agent.logger.info("event:{0} ".format(event))
                bt = self.interceptor.bt

                agent.logger.info("bt statment....{0}".format(bt))
                if not bt:
                    bt = agent.get_current_bt()
                agent.logger.info("new bt value{0}".format(bt))
                if bt:
                    backend = self.get_backend(event.connection_id, event.database_name)
                    agent.logger.info("backend..........{0}".format(backend))
                    agent.logger.info("Modulenameintercept_pymongo class ||  backend is{0}".format(backend))
                    if backend:
                        agent.logger.info("host.....".format(backend["HOST"]))

                        if backend and not hasattr(event, '_pythonagent_exit_call'):
                            exit_call = self.interceptor.db_call_begin(bt, backend["HOST"], str(event.command))
                        self.exit_call_map[event.operation_id] = exit_call

        def succeeded(self, event):
            self.interceptor.db_call_end(self.interceptor.bt,self.exit_call_map.pop(event.operation_id,None))
        def failed(self, event):
            self.interceptor.db_call_end(self.interceptor.bt,self.exit_call_map.pop(event.operation_id,None))

    mod.monitoring.register(ExitCallListener())




from __future__ import unicode_literals
import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor

class FalconExceptionAdder(BaseInterceptor):
    def add_exception(self, func, *args, **kwargs):
        with self.log_exceptions():
            bt = self.bt
            if bt:
                bt.add_exception(*sys.exc_info())
        return func(*args, **kwargs)


def intercept_falcon(agent, mod):
    WSGIInterceptor(agent, mod.App).attach('__call__')
    FalconExceptionAdder(agent, mod.App).attach('_handle_exception', patched_method_name='add_exception')
    #FalconExceptionAdder(agent, mod.Request).attach('set_response', patched_method_name='add_exception')




"""This module contains all interactions with raw Python frame objects.

This module's public functions should never ever return a raw frame object to
the caller.  To explain all the weird try: finally: del frame stuff, see:
https://docs.python.org/3.4/library/inspect.html#the-interpreter-stack.

"""

from __future__ import unicode_literals
import sys
import traceback


def is_internal_frame(filename):
    """Return True if this filename belongs to an internal CavAgent module.

    """
    return 'site-packages/CavAgent' in filename


class FrameInfo(object):
    """Copy of the important info from a Python frame.

    """
    __slots__ = ('class_name', 'name', 'filename', 'lineno', 'internal')

    def __init__(self, frame):
        try:
            try:
                self.class_name = frame.f_locals['self'].__class__.__name__
            except:
                self.class_name = None

            code = frame.f_code
            self.name = code.co_name
            self.filename = code.co_filename
            self.lineno = code.co_firstlineno
            self.internal = is_internal_frame(self.filename)
        finally:
            del frame

    def __eq__(self, rhs):
        try:
            return (
                self.name == rhs.name and
                self.filename == rhs.filename and
                self.lineno == rhs.lineno and
                self.class_name == rhs.class_name
            )
        except:
            return False

    def __ne__(self, rhs):
        return not self == rhs

    def __hash__(self):
        return hash((self.name, self.filename, self.lineno))


def _frame_hash(frame):
    try:
        code = frame.f_code
        key = (code.co_name, code.co_filename, code.co_firstlineno)
        return hash(key)
    finally:
        del frame


def _get_root_frame(thread_id):
    # Internal use only - get the root frame for the specified thread.
    frames = sys._current_frames()
    try:
        return frames.get(thread_id, None)
    finally:
        del frames


def get_formatted_stack(thread_id):
    """Return a list of formatted strings representing the current stack.

    """
    root = _get_root_frame(thread_id)
    try:
        return traceback.format_stack(root)
    finally:
        del root


def get_frame_info(level):
    """Drop in replacement for `sys._getframe` which returns a FrameInfo object.

    """
    frame = sys._getframe(level + 1)
    try:
        return FrameInfo(frame)
    finally:
        del frame


def get_stack(frame_cache, from_root=True, thread_id=None, greenlet_ref=None):
    """Generate a list of FrameInfo objects representing the python call stack.

    By default this function tries to generate a full call stack from the root
    frame on the specified thread or greenlet id.  When calling this function
    for this purpose a thread_id must *always* be specified.

    Passing `from_root=False` will generate a stack from the current frame
    outwards.  When calling in this way a thread or greenlet id is not required.

    `frame_cache` is a dict which is used to cache FrameInfo objects in order
    to reduce the memory footprint of generating stacks with many duplicate
    frames.  The management of this cache is left up to the caller.

    """

    frame = None
    try:
        if from_root:
            if greenlet_ref:
                ctx = greenlet_ref()
                if ctx:
                    # note that ctx.gr_frame may be none indicating that the current BT is not running
                    # under the context of greenlet but thread
                    frame = ctx.gr_frame or _get_root_frame(thread_id)
                else:
                    return None
            else:
                frame = _get_root_frame(thread_id)

            if frame is None:
                return None
        else:
            frame = sys._getframe()

        stack = []
        while frame:
            f_hash = _frame_hash(frame)
            try:
                stack.append(frame_cache[f_hash])
            except KeyError:
                f_info = FrameInfo(frame)
                frame_cache[f_hash] = f_info
                stack.append(f_info)

            next_frame = frame.f_back
            del frame
            frame = next_frame
            del next_frame
        return stack
    finally:
        del frame




"""Interceptor for Flask framework.

"""

from __future__ import unicode_literals
import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor


class FlaskInterceptor(BaseInterceptor):
    def _handle_user_exception(self, handle_user_exception, flask, e):
        self.agent.logger.info('Modulename: FlaskInterceptor  class')
        with self.log_exceptions():
            bt = self.bt
            self.agent.logger.info("Modulename: FlaskInterceptor  class || bt  value is {0} :".format(bt))
            if bt:
                #bt.add_exception(*sys.exc_info())
                self.agent.logger.debug("Exception occured as  !!! {0}".format(sys.exc_info()))
        #print('returning from _handle method !!')
        return handle_user_exception(flask, e)


def intercept_flask(agent, mod):
    #print('=============inside flask instrumentation==========')
    WSGIInterceptor(agent, mod.Flask).attach('wsgi_app')
    FlaskInterceptor(agent, mod.Flask).attach('handle_user_exception')



from __future__ import unicode_literals
import functools
import inspect
import itertools
import os
import sys

# The aim here is make Python 2 behave like Python 3.
# This means we write code in modern Python 3 style, where everything is an
# iterator, everything is unicode etc.

# pylint: disable=import-error,no-name-in-module,no-member,undefined-variable

SUPER_FUN_HAPPY_PYTHON = sys.version_info[0] > 2

if SUPER_FUN_HAPPY_PYTHON:
    # Imports
    # from configparser import SafeConfigParser
    # from http.client import HTTPConnection, HTTPSConnection
    # from http.cookies import SimpleCookie
    try:
        from importlib import reload
    except ImportError:
        from imp import reload
    import queue
    import _thread as thread
    from urllib.request import urlopen
    from urllib.parse import parse_qs, parse_qsl, urlparse

    # Types
    bytes_t = bytes
    long_t = int
    str_t = str

    # Functions
    filter = filter
    getcwd = os.getcwd
    items = lambda d: d.items()
    keys = lambda d: d.keys()
    long = int
    map = map
    range = range
    wraps = functools.wraps
    values = lambda d: d.values()
    zip = zip

    def get_args(callable):
        return inspect.signature(callable).parameters

    import importlib

    def import_module(name):
        return importlib.import_module(name)
else:
    # Imports
    from __builtin__ import reload
    from ConfigParser import SafeConfigParser
    from Cookie import SimpleCookie
    from httplib import HTTPConnection, HTTPSConnection
    import Queue as queue
    import thread
    from urllib import urlopen
    from urlparse import parse_qs, parse_qsl, urlparse

    # Types
    bytes_t = bytes
    long_t = long
    str_t = unicode

    # Functions
    filter = lambda f, it: itertools.ifilter(f, it)
    getcwd = os.getcwdu
    items = lambda d: d.iteritems()
    keys = lambda d: d.iterkeys()
    long = long
    map = lambda f, *it: itertools.imap(f, *it)
    range = xrange
    values = lambda d: d.itervalues()
    zip = lambda *it: itertools.izip(*it)

    def get_args(callable):
        return inspect.getargspec(callable).args

    import imp

    def import_module(name):
        file, pathname, desc = imp.find_module(name)
        try:
            return imp.load_module(name, file, pathname, desc)
        finally:
            if file is not None:
                file.close()

    # modified from https://github.com/michilu/python-functools32/blob/master/functools32/functools32.py
    def _update_wrapper(wrapper, wrapped, assigned=functools.WRAPPER_ASSIGNMENTS, updated=functools.WRAPPER_UPDATES):
        wrapper.__wrapped__ = wrapped
        for attr in assigned:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
        for attr in updated:
            getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
        return wrapper

    def wraps(wrapped, assigned=functools.WRAPPER_ASSIGNMENTS, updated=functools.WRAPPER_UPDATES):
        return functools.partial(_update_wrapper, wrapped=wrapped, assigned=assigned, updated=updated)

# Common Stuff
native_str = str
reduce = functools.reduce


def bytes(s, encoding='utf-8'):
    # Encode all bytes strings as utf-8.
    if isinstance(s, bytes_t):
        return s
    else:
        # This None thing isn't ideal, but since we use this function mainly
        # for making byte strings for protobufs, we want None to stay None.
        return None if s is None else s.encode(encoding)


def str(s, encoding='utf-8'):
    # Assume all bytes strings are utf-8 encoded.
    if isinstance(s, bytes_t):
        return s.decode(encoding)
    else:
        return str_t(s)





"""Definition of base entry and exit point interceptors.

"""


import sys
from functools import wraps
from contextlib import contextmanager
import time


class BaseInterceptor(object):
    def __init__(self, agent, cls):
        self.agent = agent
        self.cls = cls

    @property
    def bt(self):
        # add a docstring here about what current bt means now.
        return self.agent.get_current_bt()

    def __setitem__(self, key, value):
        bt = self.bt
        if bt:
            bt._properties[key] = value

    def __getitem__(self, key):
        bt = self.bt
        if bt:
            return bt._properties.get(key)

    def __delitem__(self, key):
        bt = self.bt
        if bt:
            bt._properties.pop(key, None)


    @staticmethod
    def _fix_dunder_method_name(method, class_name):
        # If `method` starts with '__', then it will have been renamed by the lexer to '_SomeClass__some_method'
        # (unless the method name ends with '__').
        if method.startswith('__') and not method.endswith('__'):
            method = '_' + class_name + method
        return method

    def _attach(self, method, wrapper_func, patched_method_name):
        patched_method_name = patched_method_name or '_' + method

        # Deal with reserved identifiers.
        # https://docs.python.org/2/reference/lexical_analysis.html#reserved-classes-of-identifiers
        method = self._fix_dunder_method_name(method, self.cls.__name__)
        patched_method_name = self._fix_dunder_method_name(patched_method_name, self.__class__.__name__)

        # Wrap the original method if required.
        original_method = getattr(self.cls, method)

        # Do not intercept the same method more than once.
        if hasattr(original_method, '_pythonagent_intercepted'):
            return

        if wrapper_func:
            @wraps(original_method)
            def wrapped_method(*args, **kwargs):
                return wrapper_func(original_method, *args, **kwargs)
            real_method = wrapped_method
        else:
            real_method = original_method

        # Replace `self.cls.method` with a call to the patched method.
        patched_method = getattr(self, patched_method_name)

        @wraps(original_method)
        def call_patched_method(*args, **kwargs):
            return patched_method(real_method, *args, **kwargs)

        call_patched_method._pythonagent_intercepted = True

        setattr(self.cls, method, call_patched_method)

    def attach(self, method_or_methods, wrapper_func=None, patched_method_name=None):
        if not isinstance(method_or_methods, list):
            method_or_methods = [method_or_methods]
        for method in method_or_methods:
            self._attach(method, wrapper_func, patched_method_name)

    def log_exception(self, level=1):
        self.agent.logger.exception('Exception in {klass}.{function}.'.format(klass=self.__class__.__name__, function=level))

    @contextmanager
    def log_exceptions(self):
        try:
            yield
        except:
            self.log_exception(level=3)


NO_WRAPPER = object()


class ExitCallInterceptor(BaseInterceptor):
    def attach(self, method_or_methods, wrapper_func=NO_WRAPPER, patched_method_name=None):
        if wrapper_func is NO_WRAPPER:
            wrapper_func = self.run
        super(ExitCallInterceptor, self).attach(method_or_methods, wrapper_func=wrapper_func,
                                                patched_method_name=patched_method_name)

    def make_correlation_header(self, exit_call):
        header = None
        if exit_call and header is not None:
            exit_call.optional_properties['CorrelationHeader'] = header[1]
        return header

    def http_call_begin(self, bt, http_host, url):
        """Start an exit call.
        """
        print("HTTP CALL BEGIN")
        current_time = time.time_ns() // 1000000
        self.agent.get_transaction_context().tier_callout_start_time = current_time
        self.agent.get_transaction_context().tier_callout_type = "http"
        self.agent.get_transaction_context().backend_header = http_host
        method_name = "request" + "." + self.agent.get_transaction_context().http_request_type.lower()
        zero_bt = 0

        with self.log_exceptions():

            self.agent.method_entry_http_callout(zero_bt, method_name, http_host, url)  # METHOD ENTRY FOR PROXY
            return self.agent.http_call_begin(bt, http_host, url)

    def db_call_begin(self, bt, db_host, db_query):
        """Start an exit call.
        """
        with self.log_exceptions():
            return self.agent.db_call_begin(bt, db_host, db_query)

    def run(self, func, *args, **kwargs):
        """Run the function.  If it raises an exception, end the exit call started from func
           and raise the exception.

           The exit call that needs to be managed should be passed as key word argument pythonagent_exit_call.

        """
        exit_call = kwargs.pop('pythonagent_exit_call', None)
        with self.end_exit_call_and_reraise_on_exception(exit_call):
            return func(*args, **kwargs)

    def http_call_end(self, bt, exit_call, status=200):
        """End the exit call.
        """
        print("HTTP CALL END")
        current_time = time.time_ns() // 1000000
        start_time = self.agent.get_transaction_context().tier_callout_start_time
        duration = current_time - start_time
        method_name = "request" + "." + self.agent.get_transaction_context().http_request_type.lower()
        zero_bt = 0

        try:
            from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException
            havoc_monitor = NDNetHavocMonitor.get_instance()
            http_host = self.agent.get_transaction_context().backend_header
            havoc_monitor.apply_outbound_service_failure(http_host)

        except Exception as e:
            if isinstance(e, NDHavocException):
                self.agent.logger.info("Havoc Exception {}".format(e))
                zero_bt = 0
                backend_header = self.agent.get_transaction_context().backend_header
                status = 503
                self.agent.method_exit_http_callout(zero_bt, method_name, backend_header, status, duration)
                self.agent.set_current_status_code(503)
                raise e
            else:
                self.agent.logger.info("Non-Havoc Exception {}".format(e))

        with self.log_exceptions():
            if exit_call:
                print("Inside with self.log_exceptions(); if exit_call")
                backend_header = self.agent.get_transaction_context().backend_header
                self.agent.method_exit_http_callout(zero_bt, method_name, backend_header, status, duration)  # METHOD EXIT FOR PROXY
                self.agent.http_call_end(bt, exit_call)


    def db_call_end(self, bt, ip_handle):
        """End the exit call.
        """
        with self.log_exceptions():
            if bt:
                self.agent.db_call_end(bt, ip_handle)

    @contextmanager
    def end_exit_call_and_reraise_on_exception(self, exit_call, ignored_exceptions=()):
        try:
            yield
        except ignored_exceptions:
            raise
        except:
            self.agent.logger.exception("Exception raised in end exit call !!! {0}".format(sys.exc_info()))
            raise


class EntryPointInterceptor(BaseInterceptor):
    HTTP_ERROR_DISPLAY_NAME = 'HTTP {code}'

    def start_business_transaction(self, bt_name, correlation_header):
        with self.log_exceptions():
            return self.agent.start_business_transaction(bt_name, correlation_header)

    def end_business_transaction(self, bt):
        with self.log_exceptions():
            self.agent.end_business_transaction(bt)

    def handle_http_status_code(self, bt, status_code, msg):
        """Add the status code to the BT and deal with error codes.

        If the status code is in the error config and enabled, or the status
        code is >= 400, create an ErrorInfo object and add it to the BT.

        """
        self.agent.set_current_status_code(status_code)
        if status_code >= 400:
            self.agent.logger.info('Message is '.format(msg))
        else:
            return







"""Utilities for Cavisson Python Agent code.

"""

from __future__ import unicode_literals
import errno

from logging import Formatter
import os

#mport _thread as thread
from .lang import thread

try:
    import greenlet
    try:
        get_ident = lambda: hash(greenlet.getcurrent())
#        print "hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh"
        GREENLETS_ENABLED = True
    except:
 #      print "get_ident except "
        get_ident = thread.get_ident().ident

   # get_ident = lambda: hash(greenlet.getcurrent())
   # GREENLETS_ENABLED = True
except ImportError:
    get_ident = thread.get_ident
    GREENLETS_ENABLED = False

items = lambda d: d.items()


default_log_formatter = Formatter('%(asctime)s [%(levelname)s] %(funcName)s(%(lineno)d) <%(process)d>: %(message)s')

class LazyWsgiRequest(object):
    """Lazily read request line and headers from a WSGI environ.

    This matches enough of the Werkzeug Request API for the agent's needs: it
    only provides access to the information in the request line and the
    headers that is needed for the agent. (Since the agent doesn't inspect
    the request body, we don't touch any of that.)

    Parameters
    ----------
    environ : dict
        A WSGI environment.

    Attributes
    ----------
    headers : dict
        A dictionary of the HTTP headers. The headers are lowercase with
        dashes separating words.
    method : str
        The request method (e.g., GET).
    url : str
        The URL of the request (reconstructed according to PEP 333).
    cookies : dict
        The cookies passed in the request header (if any).
    path : str
        The path part of the request. Note that unlike raw WSGI, this will be
        just '/' if it would otherwise be empty.
    args : dict
        The query parameters. This is not a multi-dict: if a parameter is
        repeated multiple times, one of them wins.
    referer : str
        The HTTP Referer string.
    user_agent : str
        The HTTP User-Agent string.
    is_ajax : bool
        True if this request is AJAX.
    is_mobile : bool
        True if this request is from mobile.

    """
    DEFAULT_PORTS = {
        'http': 80,
        'https': 443,
    }

    def __init__(self, environ):
        super(LazyWsgiRequest, self).__init__()
        self.environ = environ.copy()

        self._headers = None
        self._host = None
        self._port = None
        self._http_host = None
        self._url = None
        self._path = None
        self._args = None
        self._cookies = None

def mkdir(path):
    """Create the directory in path, creating any intermediate directories.

    Does not complain if the directory already exists.

    """
    try:
        os.makedirs(path)
        #print("Logging path: ",path)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise



"""Interceptors and utilities for dealing with WSGI-based apps/frameworks"""
import os
import sys
import pstats 
from sys import settrace
from datetime import datetime, timedelta
import imp
import inspect
from functools import wraps
import threading
import pythonagent.agent
from ..base import EntryPointInterceptor
#from lib import LazyWsgiRequest
from pythonagent import config
ENTRY_WSGI =8
#from agent.internal.proxy import *
from werkzeug.wrappers import Request ,Response
from pyclbr import readmodule
import os
import inspect
import cProfile
import pstats
from pythonagent.agent.probes.frameworks.test import dumpstacks
from pythonagent.agent.probes.frameworks.aisess import Aisession
#import AIsession
#import aisess
#aisess
#import importlib.util
#spec = importlib.util.spec_from_file_location("aisess", "/home/cavisson/my-shop/env/env/lib/python3.7/site-packages/pythonagent/agent/probes/frameworks/aisess.py")
#foo = importlib.util.module_from_spec(spec)
#spec.loader.exec_module(foo)

#os.environ['Aisession_flag'] = 'True'

#profiler = None
def aiflagset(s,st,d,sname):
    os.environ['Aisession_flag'] = 'True'   
    global duration
    duration = d
    global starttime
    starttime = st
    global istart
    istart = s 
    global sess
    sess = sname

class WSGIInterceptor(EntryPointInterceptor):

    def attach(self, application):
        super(WSGIInterceptor, self).attach(application, patched_method_name='application_callable')

    def application_callable(self, application, instance, environ, start_response):
        # bt = self.start_transaction(ENTRY_WSGI, LazyWsgiRequest(environ))
        stack = inspect.stack()
        temp = {}
        print("the_method in wsgi.py ------------------------>>>>>>>>>>>>",stack)
        the_method_called = stack[3][0].f_code.co_name
        print("the_method in wsgi.py ------------------>>>>>>>>>>>>>>>>>>>>>>",the_method_called)

        request = Request(environ)
        
        
        #thread dump captured
        #dumpstacks()

        # local trace function which returns itself
        # local trace function which returns itselfi

        
        if os.environ.get('Aisession_flag', 'None') == 'True' and os.environ.get('aisess_status', 'False') != 'True':
            profiler = Aisession(istart, starttime, duration, sess)
            profiler.enableprof()
            profiler.disableprof()

        bt = self.start_business_transaction(request.full_path, '')

        self.agent.logger.debug("Modulename: WSGI interceptor class || bt value is :{0}".format(bt)) 


        try:
            self.agent.logger.info("Modulename: WSGI interceptor class || request is :{0} || request path is {1} || query_string: {2} || host: {3} ".format(request,(request.path,request.query_string,request.host)))
        except:
            pass

        try:
            response = application(instance, environ, self._make_start_response_wrapper(start_response))
            self.agent.logger.info("Modulename: WSGI interceptor class || response is :{0}".format(response))

        except:
            with self.log_exceptions():
                if bt:
                    bt.add_exception(*sys.exc_info())

            raise
        finally:

            method = the_method_called
            self.end_business_transaction(bt)

        return response

    def _make_start_response_wrapper(self, start_response):
        @wraps(start_response)
        def start_response_wrapper(status, headers, exc_info=None):
            #Deal with HTTP status codes, errors and EUM correlation.
            #See https://www.python.org/dev/peps/pep-0333/#the-start-response-callable for more information.
            with self.log_exceptions():
                bt = self.bt
                self.agent.logger.info("Modulename: WSGI interceptor class || bt value is :{0}".format(bt))

                if bt:
                    # Store the HTTP status code and deal with errors.
                    status_code, msg = status.split(' ', 1)
                    self.handle_http_status_code(bt, int(status_code), msg)

                    # Inject EUM metadata into the response headers.
                    # inject_eum_metadata(self.agent.eum_config, bt, headers)

            return start_response(status, headers, exc_info)

        return start_response_wrapper


class WSGIMiddleware(object):
    #self.agent.logger.info('Modulename: WSGIMiddleware class ')
    def __init__(self, application=None):
        self._application = application
        self._configured = False
        self._interceptor = WSGIInterceptor(pythonagent.agent.get_agent_instance(), None)
        self._load_application_lock = threading.Lock()
        #get_script()
        #from pythonagent import config
        #config.WSGI_SCRIPT_ALIAS = input("Enter the absolute path of WSGI application script. (command to get 'readlink -f <script_path.py>') : ")

    def load_application(self):
        wsgi_callable = config.WSGI_CALLABLE_OBJECT or 'application'

        config.WSGI_SCRIPT_ALIAS = os.environ.get('WSGI_SCRIPT_ALIAS')
        if not config.WSGI_SCRIPT_ALIAS and not config.WSGI_MODULE:
            raise AttributeError(
                'Cannot get WSGI application: the agent cannot load your '
                'application. You must set either CAV_WSGI_SCRIPT_ALIAS'
                'in order to load your application.')

        if config.WSGI_MODULE:
            module_name = config.WSGI_MODULE

            if ':' in module_name:
                module_name, wsgi_callable = module_name.split(':', 1)

            __import__(module_name)
            wsgi_module = sys.modules[module_name]
        else:
            wsgi_module = imp.load_source('wsgi_module', config.WSGI_SCRIPT_ALIAS)

        if wsgi_callable.endswith('()'):  # "Quick" callback
            app = getattr(wsgi_module, wsgi_callable[:-2])
            app = app()
        else:
            app = getattr(wsgi_module, wsgi_callable)

        self._application = app

    def wsgi_application(self, environ, start_response):
        return self._application(environ, start_response)

    def __call__(self, environ, start_response):
        #config.WSGI_SCRIPT_ALIAS = input("Enter the absolute path of WSGI application script. (command to get 'readlink -f <script_path.py>') : ")
        if not self._configured:
            pythonagent.agent.configure(environ)
            self._configured = True

        if not self._application:
            # CORE-60212 - We are deadlocking inside imp.load_source when
            # WSGI_SCRIPT_ALIAS is used and uwsgi workers are restarted. We
            # actually don't need to go inside `load_application` multiple
            # times, which should prevent the imp.load_source deadlock while
            # improving initial request performance a bit. (We do double-
            # checked locking so we don't take the lock on every __call__,
            # only the initial burst of concurrent ones.)
            with self._load_application_lock:
                if not self._application:
                    self.load_application()

        # The interceptor expects an unbound function to call, hence why this function is called like this.
        return self._interceptor.application_callable(WSGIMiddleware.wsgi_application, self, environ, start_response)

    def get_script():
        config.WSGI_SCRIPT_ALIAS = input("Enter the absolute path of WSGI application script. (command to get 'readlink -f <script_path.py>') : ")
        



"""Trick Python into loading the agent.

"""

from __future__ import unicode_literals
import logging
import os
import sys

try:
    sys.path.remove(os.path.dirname(__file__))
except ValueError:  # directory not in sys.path
    pass

try:
    import pythonagent.agent as agent
    #agent.Agent.start()
    agent.configure()
    agent.bootstrap()
except:
    logger = logging.getLogger('pythonagent.agent')
    logger.exception('Exception in agent startup.')
finally: 
    pass
    #appdynamics.agent.load_sitecustomize()



from __future__ import unicode_literals
import logging
from pythonagent.lib import get_ident
import ctypes
import os

from pythonagent.agent.probes.havoc.havoc_manager import NDNetHavocMonitor, NDHavocException

from pythonagent.agent.internal import udp
from pythonagent.agent.probes.frameworks import wsgi  # aiflagset
from pythonagent.agent.probes.Instrumentation.find import find



class TransactionContext(object):
    def __init__(self):
        self.nd_init_done = None
        self.request_id = None
        self.api_request_id = None
        self.aws_request_id = None
        self.function_name = None
        self.bt_id = None
        self.tier_callout_type = None
        self.tier_callout_start_time = None
        self.backend_header = None
        self.http_request_type = None
        self.tier_name = None
        self.server_name = None
        self.instance_name = None
        self.url_path = None


class Agent(object):
    """The entry point for the Python agent.

    """

    def __init__(self):
        super(Agent, self).__init__()
        self.logger = logging.getLogger('pythonagent.agent')
        self.aiobj = None
        # if os.environ['CAV_RUNNING_MODE']=='0':
        #    os.environ['CAV_RUNNING_MODE']='1'
        #    return
        self.app_id = None
        self.tier_id = None
        self.node_id = None
        self.account_guid = None
        self.controller_guid = None

        self._tx_factory = None

        # Services
        # self.proxy_control_svc = None
        # self.config_svc = None
        # self.tx_svc = None
        # self.snapshot_svc = None
        
        self.active_bts = set()
        self.current_bts = {}
        self.transaction_context = {}
        self.eum_config = None

        self.nd_init_done = None
        self.last_forced_snapshot = 0
        self.lib = None

        if "CAV_APP_AGENT_Env" in os.environ:
            if os.environ['CAV_APP_AGENT_Env'] == "AWS_LAMBDA":
                self.cav_env = "AWS_LAMBDA"
            elif os.environ['CAV_APP_AGENT_Env'] == "NATIVE":
                self.cav_env = "NATIVE"
        else:
            self.cav_env = "AWS_LAMBDA"

        if "CAV_APP_AGENT_Env" in os.environ:
            if os.environ['CAV_APP_AGENT_Env'] == "AWS_LAMBDA":
                self.cav_env = "AWS_LAMBDA"
            elif os.environ['CAV_APP_AGENT_Env'] == "NATIVE":
                self.cav_env = "NATIVE"
        else:
            self.cav_env = "AWS_LAMBDA"

        if "CAV_APP_AGENT_TIER" in os.environ:
                self.tier_name = os.environ.get("CAV_APP_AGENT_TIER")
        else:
            self.tier_name = "default"

        if "CAV_APP_AGENT_SERVER" in os.environ:
                self.server_name = os.environ.get("CAV_APP_AGENT_SERVER")
        else:
            self.server_name = "default"

        if "CAV_APP_AGENT_INSTANCE" in os.environ:
                self.instance_name = os.environ.get("CAV_APP_AGENT_INSTANCE")
                self.function_name = os.environ.get("CAV_APP_AGENT_INSTANCE")
        else:
            self.instance_name = "default"
            self.function_name = "default"


        self.start()

        if self.cav_env == "NATIVE":
            self.ai_wrapper_begin()
            self.ad_wrapper()
        # self.thd_wrapper()
        # self.ai_wrapper_end()
        # self.ai_wrapper_begin()

        self.current_status_code = {}

        self.havoc_monitor = NDNetHavocMonitor.get_instance()

    def start(self):
        path = '/usr/local/lib/libndsdk.so'

        if self.cav_env == "NATIVE":
            self.lib = self.load_common_lib(path)

        #self.logger.info("going to start sdk_init()")
        self.sdk_init()
        #self.logger.info("sdk_init() done")
        # self.nd_init_done = True
        os.environ['nd_init_done'] = '1'


    def stop(self):
        """Stop the agent from doing anything else.

        Ideally this will stop any interceptors from doing anything, it will
        disconnect all ties to the proxy etc etc.

        """
        self.nd_init_done = False
        os.environ['nd_init_done'] = '0'

        # @staticmethod
        def getInstance(self):
            """Static Access Method"""
            if self.start.__shared_instance is None:
                # Proxy()
                self.start()
            # return Proxy.__shared_instance
            return self.start.__shared_instance

    # @profile
    def load_common_lib(self, lib_path):

        self.lib = ctypes.cdll.LoadLibrary(lib_path)
        # self.lib.nd_init2.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        self.lib.nd_method_entry.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
        self.lib.nd_method_exit.argtypes = [ctypes.c_void_p, ctypes.c_char_p]

        self.lib.nd_bt_begin.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
        self.lib.nd_bt_begin.restype = ctypes.c_void_p

        self.lib.nd_bt_end.argtypes = [ctypes.c_void_p, ctypes.c_int]
        self.lib.nd_bt_end.restype = ctypes.c_int

        self.lib.nd_bt_store.argtypes = [ctypes.c_void_p, ctypes.c_char_p]

        self.lib.nd_ip_db_callout_begin.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
        self.lib.nd_ip_db_callout_begin.restype = ctypes.c_void_p

        self.lib.nd_ip_db_callout_end.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
        self.lib.nd_ip_db_callout_end.restype = ctypes.c_int

        self.lib.nd_ip_http_callout_begin.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
        self.lib.nd_ip_http_callout_begin.restype = ctypes.c_void_p

        self.lib.nd_ip_http_callout_end.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
        self.lib.nd_ip_http_callout_end.restype = ctypes.c_int

        self.lib.handleUploadDownload.argtypes = [ctypes.c_char_p]
        self.lib.handleUploadDownload.restype = ctypes.c_void_p

        # self.lib.register_AI_end_callback.argtypes = [ctypes.c_char_p]

        CB_FTYPE_DOUBLE_DOUBLE = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_int, ctypes.c_longlong,
                                                  ctypes.c_longlong, ctypes.c_char_p)
        self.c_startaisession = CB_FTYPE_DOUBLE_DOUBLE(self.startaisession)

        ai_end_ftype = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_char_p)
        self.c_endaisession = ai_end_ftype(self.endaisessioncc)

        CB_FTYPE_DOUBLE_DOUBLE = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int,
                                                  ctypes.c_int)
        self.c_startad = CB_FTYPE_DOUBLE_DOUBLE(self.thd_funct)

        thd_ftype = ctypes.CFUNCTYPE(ctypes.c_char_p)
        self.c_thd = thd_ftype(self.thd)

        self.logger.info("load common lib -> lib.obj :{0}".format(self.lib))
        return self.lib

    def ai_wrapper_begin_native(self):
        print("************************ callback registered ********************************************")
        self.lib.register_AI_start_callback(self.c_startaisession)

    def ai_wrapper_begin(self):
        if self.cav_env == "NATIVE":
            self.ai_wrapper_begin_native()
        elif self.cav_env == "AWS_LAMBDA":
            udp.ai_wrapper_begin()

    def ai_wrapper_end_native(self):
        self.lib.register_AI_end_callback(self.c_endaisession)

    def ai_wrapper_end(self):
        if self.cav_env == "NATIVE":
            self.ai_wrapper_end_native()
        elif self.cav_env == "AWS_LAMBDA":
            udp.ai_wrapper_end()

    def endaisessioncc(self, char):
        print("INSIDE END AI SESSION")

    def startaisession(self, istart, startTime, duration, sname):
        print("session status \n", istart)
        print(" IN startaisession", sname.decode('ascii'))
        wsgi.aiflagset(istart, startTime, duration, sname)
        pathd = "/opt/cavisson/netdiagnostics/logs/"
        filenameai1 = str(self.sess.decode('ascii')) + ".txt"
        aifile_to_open_to_d = os.path.join(pathd, filenameai1)
        self.lib.handleUploadDownload(aifile_to_open_to_d)


    def ad_wrapper(self):
        self.lib.register_AD_callback(self.c_startad)

    def thd_wrapper(self):
        self.lib.register_thread_dump_callback(self.c_thd)


    def thd(self):
        print("INSIDE THD")
        # dumpstacks()

    def thd_funct(self, a, b, c, d):
        find("/home/cavisson/shop/my-shop/FlaskwithMysqlapp")

    # @profile
    def sdk_init_native(self):
        self.logger.info("sdk_init -> lib.obj :{0}".format(self.lib))
        self.lib.nd_init("python", "cavlib")

    def sdk_init(self):
        if self.cav_env == "NATIVE":
            self.sdk_init_native()
        elif self.cav_env == "AWS_LAMBDA":
            udp.sdk_init(self)

    def sdk_free_native(self):
        self.logger.info("sdk_free -> lib.obj :{0}".format(self.lib))
        self.lib.nd_free()

    def sdk_free(self):
        if self.cav_env == "NATIVE":
            self.sdk_free_native()
        elif self.cav_env == "AWS_LAMBDA":
            udp.sdk_free()


    # @profile
    def method_entry_native(self, bt, method):
        print('Inside method entry ', method, bt)
        self.logger.info("method entry -> lib.obj :{0}".format(self.lib))
        self.logger.info("method entry -> method name :{0}".format(method))
        # defer C.free(unsafe.Pointer(method_c))
        # method_bytes = bytes(method, 'utf-8')
        try:
            method_bytes = bytes(method, 'utf-8')
        except:
            method_bytes = bytes(method.encode("utf-8"))

        self.lib.nd_method_entry(bt, ctypes.c_char_p(method_bytes))

    def method_entry(self, bt, method):
        if self.cav_env == "NATIVE":
            self.method_entry_native(bt, method)
        elif self.cav_env == "AWS_LAMBDA":
            udp.method_entry(self, bt, method, "NA", "NA")

    def method_entry_http_callout(self, bt, method, query_string, url_parameter):

        self.havoc_monitor.apply_outbound_service_delay(query_string)

        if self.cav_env == "NATIVE":
            pass
        elif self.cav_env == "AWS_LAMBDA":
            udp.method_entry(self, bt, method, query_string, url_parameter)

    #@profile
    def method_exit_native(self, bt, method):
            print('in method exit', method, bt)
            self.logger.info("load common lib -> lib.obj :{0}".format(self.lib))
            self.logger.info("method entry -> method name :{0}".format(method))
            # defer C.free(unsafe.Pointer(method_c))
            # method_bytes = bytes(method, 'utf-8')
            try:
                method_bytes = bytes(method, 'utf-8')
            except:
                method_bytes = bytes(method.encode("utf-8"))
            print("\n\n", method_bytes)
            self.lib.nd_method_exit(bt, ctypes.c_char_p(method_bytes))

    def method_exit(self, bt, method, status=200, duration=0):
        if self.cav_env == "NATIVE":
            self.method_exit_native(bt, method)
        elif self.cav_env == "AWS_LAMBDA":
            udp.method_exit(self, bt, method, "NA", status, duration)


    def method_exit_http_callout(self, bt, method, backend_header, status, duration):
        if self.cav_env == "NATIVE":
            pass
        elif self.cav_env == "AWS_LAMBDA":
            #print("METHOD EXIT HTTP CALLOUT")
            #print("BT in agent.py ::", bt)
            #print("METHOD in agent.py ::", method)
            #print("backend_header string in agent.py ::", backend_header)
            #print("status in agent.py ::", status)
            #print("duration string in agent.py ::", duration)
            #udp.method_exit(self, bt, method)
            udp.method_exit(self, bt, method, backend_header, status, duration)
            # create_method_exit_message(context, )



    # @profile
    def start_business_transaction_native(self, bt_name, correlation_header):
        if os.environ['nd_init_done'] == '0':
            self.start()
            print("start_business_transaction ->")
            self.logger.debug(" start_business_transaction -> lib.obj {0}:".format(self.lib))
            # btname = bytes(bt_name, 'utf-8')
            # self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
            # bt_name_c = ctypes.c_char_p(btname)
            # self.logger.info("bt_name_c value: {0}".format(bt_name_c))
            # correlation_header = bytes(correlation_header, 'utf-8')
            # correlation_header_c = ctypes.c_char_p(correlation_header)
        try:
            # self.logger.debug(" start_business_transaction -> lib.obj {0}:".format(self.lib))
            btname = bytes(bt_name, 'utf-8')
            # btname = bytes(bt_name.encode("utf-8"))
            self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
            bt_name_c = ctypes.c_char_p(btname)
            self.logger.info("bt_name_c value: {0}".format(bt_name_c))
            correlation_header = bytes(correlation_header, 'utf-8')
            # correlation_header = bytes(correlation_header.encode("utf-8"))
            correlation_header_c = ctypes.c_char_p(correlation_header)
        except:
            btname = bytes(bt_name.encode("utf-8"))
            self.logger.info("btname type and value: {0} and {1}".format(type(btname), btname))
            bt_name_c = ctypes.c_char_p(btname)
            self.logger.info("bt_name_c value: {0}".format(bt_name_c))
            correlation_header = bytes(correlation_header.encode("utf-8"))
            correlation_header_c = ctypes.c_char_p(correlation_header)

        bt = self.lib.nd_bt_begin(bt_name_c, correlation_header_c)
        #self.logger.info("bt datatype : {0}".format(type(bt)))
        print("btname value: {0}".format(btname))
        # print('-----------------\nCurrent heap memory stack for BT: ',bt_name,' is : ',hpy().heap())
        self.active_bts.add(bt)
        self.set_current_bt(bt)
        print("BT value of this txn: ", bt)
        return bt

    def start_business_transaction(self, bt_name, correlation_header):
        bt = None


        """
        try:
            self.havoc_monitor.apply_inbound_service_delay(bt_name)
        except Exception as e:
            # Will fail first time
            print(e)
            pass
        """

        if self.cav_env == "NATIVE":
            bt = self.start_business_transaction_native(bt_name, correlation_header)
        elif self.cav_env == "AWS_LAMBDA":
            bt = udp.start_business_transaction(self, bt_name, correlation_header)



        try:
            self.havoc_monitor.apply_inbound_service_delay(bt_name)
        except Exception as e:
            # Will fail first time
            print(e)
            pass

        return bt

    # @profile
    def end_business_transaction_native(self, bt):
        print("end_business_transaction--> BT: ", bt)
        self.logger.info('end_business_transaction--> lib variable in end txn: {0}'.format(self.lib))
        self.logger.info('At end_business_transaction bt value is {0}'.format(bt))
        statuscode = self.get_current_status_code()
        rc = self.lib.nd_bt_end(bt, statuscode)
        self.logger.info("nd_bt_end return value:  {0}".format(int(rc)))
        self.unset_current_bt()
        self.unset_current_status_code()
        self.active_bts.discard(bt)
        return int(rc)

    def end_business_transaction(self, bt):
        if self.cav_env == "NATIVE":
            rc = self.end_business_transaction_native(bt)
        elif self.cav_env == "AWS_LAMBDA":
            rc = udp.end_business_transaction(self, bt)
        return rc


    #@profile
    def store_business_transaction_native(self, bt, unique_bt_id):
        self.logger.info('lib variable in store business txn: {0}'.format(self.lib))
        # unique_bt_id_bytes = bytes(unique_bt_id, 'utf-8')
        try:
            unique_bt_id_bytes = bytes(unique_bt_id, 'utf-8')
        except:
            unique_bt_id_bytes = bytes(unique_bt_id.encode("utf-8"))

        bt_id_c = ctypes.c_char_p(unique_bt_id_bytes)
        self.lib.nd_bt_store(bt, bt_id_c)


    def store_business_transaction(self, bt, unique_bt_id):
        if self.cav_env == "NATIVE":
            self.store_business_transaction_native(bt, unique_bt_id)
        elif self.cav_env == "AWS_LAMBDA":
            udp.store_business_transaction(bt, unique_bt_id)


    #@profile
    def db_call_begin_native(self, bt, db_host, db_query):
        #db_host_c = bytes(db_host, 'utf-8')
        #db_query_c = bytes(db_query, 'utf-8')
        try:
            db_host_c = bytes(db_host, 'utf-8')
            db_query_c = bytes(db_query, 'utf-8')
        except:
            db_host_c = bytes(db_host.encode("utf-8"))
            db_query_c = bytes(db_query.encode("utf-8"))

        ip_handle = self.lib.nd_ip_db_callout_begin(bt, ctypes.c_char_p(db_host_c), ctypes.c_char_p(db_query_c))
        return ip_handle


    def db_call_begin(self, bt, db_host, db_query):
        if self.cav_env == "NATIVE":
            self.db_call_begin_native(bt, db_host, db_query)
        elif self.cav_env == "AWS_LAMBDA":
            udp.db_call_begin(bt, db_host, db_query)


    #@profile    #@staticmethod
    def db_call_end_native(self, bt, ip_handle):
            rc = self.lib.nd_ip_db_callout_end(bt, ip_handle)
            return int(rc)


    def db_call_end(self, bt, ip_handle):
        if self.cav_env == "NATIVE":
            self.db_call_end_native(bt, ip_handle)
        elif self.cav_env == "AWS_LAMBDA":
            udp.db_call_end(bt, ip_handle)

    #@profile  #@staticmethod
    def http_call_begin_native(self, bt, http_host, url):
        try:
            host_bytes = bytes(http_host, 'utf-8')
            url_bytes = bytes(url, 'utf-8')
        except:
            host_bytes = bytes(http_host.encode("utf-8"))
            # url_bytes = bytes(url, 'utf-8')
            url_bytes = bytes(url.encode("utf-8"))

        handle = self.lib.nd_ip_http_callout_begin(bt, ctypes.c_char_p(host_bytes), ctypes.c_char_p(url_bytes))
        return handle

    def http_call_begin(self, bt, http_host, url):
        if self.cav_env == "NATIVE":
            handle = self.http_call_begin_native(bt, http_host, url)
        elif self.cav_env == "AWS_LAMBDA":
            handle = udp.http_call_begin(self, bt, http_host, url)
        return handle

    #@profile #@staticmethod
    def http_call_end_native(self, bt, ip_handle):
        rc = self.lib.nd_ip_http_callout_end(bt, ip_handle)
        return int(rc)

    def http_call_end(self, bt, ip_handle):
        if self.cav_env == "NATIVE":
            self.http_call_end_native(bt, ip_handle)
        elif self.cav_env == "AWS_LAMBDA":
            udp.http_call_end(self, bt, ip_handle)

    def wait_for_start(self, timeout_ms=None):
            """Wait for the agent to start and get configured.

            Other Parameters
            ----------------
            timeout_ms : int, optional
                The maximum time to wait for the agent to start and be configured
                before returning.

            Returns
            -------
            bool
                Returns ``True`` if the agent is enabled after waiting, else
                ``False``.

            """
            if timeout_ms is not None:
                with self.timer() as timer:
                    if self.proxy_control_svc is not None:
                        self.proxy_control_svc.wait_for_start(timeout_ms=timeout_ms)

                    if self.config_svc is not None:
                        timeout_ms = max(0, timeout_ms - timer.duration_ms)
                        self.config_svc.wait_for_config(timeout_ms=timeout_ms)

                    if self.tx_svc is not None:
                        timeout_ms = max(0, timeout_ms - timer.duration_ms)
                        self.tx_svc.wait_for_start(timeout_ms=timeout_ms)
            else:
                self.proxy_control_svc.wait_for_start()
                self.config_svc.wait_for_config()
                self.tx_svc.wait_for_start()

            return self.enabled


    def wait_for_end(self, timeout_ms=None):
        """Wait for the agent to finish reporting any pending BTs.

        """
        self.tx_svc.wait_for_end(timeout_ms=timeout_ms)

    #@property
    def enabled(self):
            """Return true if the agent has started and is enabled.

            """
            return (
                    self.nd_init_done and
                    self.config_svc and self.config_svc.enabled and
                    self.tx_svc)

    def get_current_bt(self):
        """Get the currently active BT for the calling context.

        The calling context is the active greenlet or thread, depending on
        whether greenlets are in use or not. If the agent is disabled, or if
        there is no active transaction for the calling context, None is
        returned.

        Returns
        -------
        bt : appdynamics.agent.core.bt.Transaction or None
            the active business transaction (if any)

        """
        # if not self.enabled:
        #    return None
        return self.current_bts.get(get_ident(), None)

    def set_current_bt(self, bt):
        if bt:
            self.current_bts[get_ident()] = bt

    def unset_current_bt(self):
        self.current_bts.pop(get_ident(), None)

    def get_current_status_code(self):
        return self.current_status_code.get(get_ident(), None)

    def set_current_status_code(self, status_code):
        if status_code:
            self.current_status_code[get_ident()] = status_code

    def unset_current_status_code(self):
        self.current_status_code.pop(get_ident(), None)
    
    def set_transaction_context(self, context):

        #print("INVOKED set_transaction_context ", context)
        if context:
            context.tier_name = self.tier_name
            context.server_name = self.server_name
            context.instance_name = self.instance_name
            context.function_name = self.function_name
            self.transaction_context[get_ident()] = context
    
    def get_transaction_context(self):
        context = self.transaction_context.get(get_ident(), None)
        if context is None:
            context = TransactionContext()
            self.set_transaction_context(context)
        return context

    def reset_transaction_context(self):
        self.transaction_context.pop(get_ident(), None)
        self.current_status_code.pop(get_ident(), None)
        self.current_bts.pop(get_ident(), None)


class AgentNotStartedException(Exception):
    pass


class AgentNotReadyException(Exception):
    pass


class IgnoreTransaction(Exception):
    pass









"""Interceptor for Pyramid.

"""


from __future__ import unicode_literals
import sys

from pythonagent.agent.probes.frameworks.wsgi import WSGIInterceptor
from pythonagent.agent.probes.base import BaseInterceptor

class PyramidRouterInterceptor(BaseInterceptor):
    """
    Pyramid's interceptor class
    """
    def _handle_request(self, handle_request, router_instance, req):
        """

        Parameters
        ----------
        handle_request: original method pyramid.router.Router.handle_request
                        that is being overridden
        router_instance: pyramid.router.Router instance
        req: request object

        Returns
        -------
        Whatever pyramid.router.Router.handle_request returns

        """
        try:
            return handle_request(router_instance, req)
        except Exception:
            with self.log_exceptions():
                bt = self.bt
                if bt:
                    bt.add_exception(*sys.exc_info())
            raise


def intercept_pyramid(agent, mod):
    """

    Parameters
    ----------
    
    mod: pyramid.router module

    Returns
    -------
    None
    """
    #mod = '/home/cavisson/shop/my-shop/env/env/lib/python3.7/site-packages/pyramid/router.py'
    WSGIInterceptor(agent, mod.Router).attach('__call__')
    PyramidRouterInterceptor(agent, mod.Router).attach('handle_request') 




"""Intercept urllib3 to ensure that HTTPS is reported correctly.

"""

from __future__ import unicode_literals

from . import HTTPConnectionInterceptor


def intercept_urllib3(agent, mod):
    # urllib3 1.8+ provides its own HTTPSConnection class.
    if hasattr(mod, 'connection'):
        #print("intercept_urllib3")
        HTTPConnectionInterceptor.https_connection_classes.add(mod.connection.HTTPSConnection)



import sqlite3
con = sqlite3.connect('checklist.db')
con.execute("CREATE TABLE checklist(id INTEGER PRIMARY KEY,task VARCHAR(100) NOT NULL,description VARCHAR(200) NOT NULL,status VARCHAR(3) NOT NULL)")
con.execute("INSERT INTO checklist(id,task,description,status) VALUES (1,'Task number 1', 'This is a description about Task number 1', 'Yes')")
con.execute("INSERT INTO checklist(id,task,description,status) VALUES (2,'Task number 2', 'This is a description about Task number 2', 'Yes')")
con.execute("INSERT INTO checklist(id,task,description,status) VALUES (3,'Task number 3', 'This is a description about Task number 3', 'Yes')")
con.commit() 




from pythonagent.find import find
import traceback
import sys

py_ver = sys.version_info[0]

USAGE = "auto discovery to generate instrumentation profile"
ABOUT = "Command to push auto discovery feature of Python Agent"


def command(options, args):
    # Calling auto discovery method to discover and create instrumentation profile
    #print(args[0])
    #if args[0] is None:
    #    raise TypeError("'None' value provided for application path !!")   
    try: 
        find(args[0])
        print('auto discovery called, Instrumentation profile has been generated !!')
    except:
        if py_ver == 3: 
            traceback.print_exc(limit=None, file=None, chain=True)
        else:
            print_exception(sys.exc_type, sys.exc_value, sys.exc_traceback, limit, file)
            print('No path provided for application !!\nPlease provide application path along with commond.')
        



import sys
import os
import ast
import json

#path ="C:/Users/garima.singh/PycharmProjects/django_redis_demo"
path ="/home/cavisson/shop/my-shop/pythonagent_garima"

modules = []
default_path = []

for root, dirs, files  in os.walk(path):
    for file in files:
        if(file.endswith(".py")):
            modules.append(file)
            default_path.append(os.path.join(root,file))
           
print("modules",modules)
print("default_path",default_path)


data = {
  "modules": [
    {
      "moduleName": filename,
      "moduleLevelMethod": [
      {
        "name":fname.name ,
        "signature":"NA"
      }
      for fname in [n for n in ast.parse(open(filename).read()).body if isinstance(n, ast.FunctionDef)]
      ],
      "classes": [
        {
          "className": cname.name,
          "classLevelMethods": [
      {
        "name":clmname.name,
        "signature":"NA"
     
      } 
        for clmname in [n for n in cname.body if isinstance(n, ast.FunctionDef)]      
      ]
        }
        for cname in [n for n in ast.parse(open(filename).read()).body if isinstance(n, ast.ClassDef)]
      ]
    }
    for filename in default_path
   ]
   }

json_object = json.dumps(data, indent = 2) 
with open("sample_json.json", "w") as outfile: 
    outfile.write(json_object)





"""Intercept the psycopg2 package.

The interceptors in here are more complex than for other packages because the
classes are defined in C extension code.  We have to wrap the connection and
cursor classes with our own, which allows us to attach to the relevant methods.

"""

# pylint: disable=no-self-argument

from __future__ import unicode_literals
import re

from pythonagent.lang import parse_qs, str, urlparse
from ..base import BaseInterceptor
from .dbapi import DbAPIConnectionInterceptor, DbAPICursorInterceptor

# For format of PostgreSQL connection strings, see:
# http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-CONNSTRING

KV_RE = re.compile(r'\s*(?P<key>[a-zA-Z_]+)\s*=\s*(?P<remainder>.+)')


def parse_postgresql_url_dsn(dsn):
    url = urlparse(dsn, allow_fragments=False)

    if '?' in url.path:
        parts = url.path.split('?', 1)
        path, params = parts[0], parse_qs(parts[1])
    else:
        path = url.path
        params = parse_qs(url.query) if url.query else {}

    if 'host' in params:
        host = params['host'][0]
    else:
        host = url.hostname or 'localhost'

    if 'port' in params:
        port = params['port'][0]
    elif url.port:
        port = str(url.port)
    else:
        port = '5432'

    path = path.strip('/')

    if 'dbname' in params:
        dbname = params['dbname'][0]
    elif path:
        dbname = path
    elif 'user' in params:
        dbname = params['user'][0]
    else:
        dbname = url.username or 'postgres'

    return host, port, dbname


def parse_quoted_value(remainder):
    length = len(remainder)
    idx = 1

    start_idx = idx
    val = []

    while idx < length and remainder[idx] != "'":
        if remainder[idx:idx + 2] in ("\\\\", "\\'"):
            # Valid escape sequence: collect the current run with the
            # escaped character, move past the escape sequence, and start a
            # new run.
            val.append(remainder[start_idx:idx] + remainder[idx + 1])
            idx += 2
            start_idx = idx
        else:
            idx += 1

    val.append(remainder[start_idx:idx])
    return ''.join(val), remainder[idx + 1:]


def parse_postgresql_kv_dsn(dsn):
    if "'" not in dsn:  # Fast path when there are no quoted strings.
        dsn = re.sub(r'(\s*=\s+|\s+=\s*)', '=', dsn).split()
        parts = dict(pair.split('=', 1) for pair in dsn)
    else:
        parts = {}
        m = KV_RE.match(dsn)

        while m:
            key, remainder = m.group('key'), m.group('remainder')

            if remainder[0] == "'":
                value, remainder = parse_quoted_value(remainder)
            else:
                split = remainder.split(None, 1)
                if len(split) == 1:
                    value, remainder = split[0], ''
                elif split:
                    value, remainder = split
                else:
                    break

            parts[key] = value
            m = KV_RE.match(remainder)

    return parse_postgresql_keyword_args(parts)


def parse_postgresql_keyword_args(parts):
    host = parts.get('host', 'localhost')
    port = parts.get('port', '5432')
    dbname = parts.get('dbname', parts.get('user', 'postgres'))
    return host, port, dbname


def parse_postgresql_dsn(dsn=None, *args, **kwargs):
    if dsn:
        if dsn.startswith('postgresql://'):
            return parse_postgresql_url_dsn(dsn)
        else:
            return parse_postgresql_kv_dsn(dsn)
    else:
        return parse_postgresql_keyword_args(kwargs)


class Psycopg2ConnectionInterceptor(DbAPIConnectionInterceptor):
    def get_backend_properties(self, conn, *args, **kwargs):
        return parse_postgresql_dsn(*args, **kwargs) + ('POSTGRESQL',)


class Psycopg2CursorInterceptor(DbAPICursorInterceptor):
    def get_connection(self, cursor):
        return cursor.connection


class Psycopg2Interceptor(BaseInterceptor):
    cursor_classes = {}

    def _connect(self, connect, *args, **kwargs):
        connection_factory = kwargs.pop('connection_factory', None) or self.cls.extensions.connection

        class Connection(connection_factory):
            def __init__(conn, *args, **kwargs):
                super(Connection, conn).__init__(*args, **kwargs)

            def cursor(conn, *args, **kwargs):
                cursor_factory = kwargs.pop('cursor_factory', None) or conn.cursor_factory or self.cls.extensions.cursor
                try:
                    # we store these wrapper classes to avoid 're-instrumenting' each time
                    cursor_class = self.cursor_classes[cursor_factory]
                except KeyError:
                    class Cursor(cursor_factory):
                        def execute(curs, *args, **kwargs):
                            return super(Cursor, curs).execute(*args, **kwargs)

                        def executemany(curs, *args, **kwargs):
                            return super(Cursor, curs).executemany(*args, **kwargs)
                    self.cursor_classes[cursor_factory] = Cursor
                    cursor_class = Cursor
                return super(Connection, conn).cursor(cursor_factory=cursor_class, *args, **kwargs)

        Psycopg2ConnectionInterceptor(
            self.agent, Connection, Psycopg2CursorInterceptor).attach('__init__')

        return connect(connection_factory=Connection, *args, **kwargs)


def intercept_psycopg2_connection(agent, mod):
    Psycopg2Interceptor(agent, mod).attach('connect')



"""Base class for HTTP connection interceptors.

"""

from __future__ import unicode_literals

from ..base import ExitCallInterceptor
from pythonagent.lang import str, urlparse


EXIT_HTTP =0
EXIT_SUBTYPE_HTTP = 'HTTP'

class HTTPConnectionInterceptor(ExitCallInterceptor):
    # If the library you are intercepting has an HTTPSConnection class which
    # does not subclass httplib.HTTPSConnection, add it to this set.
    https_connection_classes = set()
    backend_name_format_string = '%s://{HOST}:{PORT}{URL}?{QUERY STRING}'

    @classmethod
    def _request_is_https(cls, connection):
        #print("INSIDE IS REQUEST HTTPS")
        if connection.port == 443:
            return True
        return isinstance(connection, tuple(cls.https_connection_classes))

    def get_backend(self, host, port, scheme, url):
        #self.agent.logger.info('Modulename: HTTPConnectionInterceptor class get_backend function')
        parsed_url = urlparse(url)
        backend_properties = {
            'HOST': host,
            'PORT': str(port),
            'URL': parsed_url.path,
            'QUERY STRING': parsed_url.query,
        }
        return self.agent.backend_registry.get_backend(EXIT_HTTP, EXIT_SUBTYPE_HTTP, backend_properties,
                                                       self.backend_name_format_string % scheme)


from .httplib import intercept_httplib
from .urllib3 import intercept_urllib3
from .requests import intercept_requests
from .tornado_httpclient import intercept_tornado_httpclient
from .boto import intercept_boto

__all__ = ['intercept_httplib', 'intercept_urllib3', 'intercept_requests', 'intercept_tornado_httpclient', 'intercept_boto']

